from __future__ import annotations
import traceback
import warnings
warnings.filterwarnings("ignore")
import sys, emcee, os, time, inspect, inspect, contextlib
from emcee import EnsembleSampler
import numpy as np
from math import log10, floor
from scipy.interpolate import interp1d
import multiprocessing as mp
from beartype import beartype
from contextlib import nullcontext
from . import utils, mcmc
from .lsd import LSD
from .result import Result
from .data import Data, Config, MaskingLines, LineList, DataList
from .errors import ContinuumError
from .utils import IntLike, Scalar, Array1D, Array2D
try:
    import dynesty
except ImportError:
    dynesty = None

@beartype
class Acid:
    """
    Accurate Continuum fItting and Deconvolution (ACID) class. This class contains the ACID method 
    which fits the continuum of spectra and performs Least Squares Deconvolution (LSD) to obtain
    LSD profiles for each spectrum. It also contains many internal methods used within the main ACID 
    method. See Dolan et al (2024) for more details on the ACID method and its applications.
    """

    def __init__(
        self,
        velocities       : Array1D|None                                       = None,   # Data
        linelist         : Array2D|str|LineList|dict|None                     = None,   # Data
        order            : IntLike|None                                       = None,   # Config
        order_range      : Array1D|None                                       = None,   # Config
        verbose          : IntLike|bool|str|None                              = None,   # Config
        sampler_progress : bool|None                                          = None,   # Config
        masking_lines    : dict|MaskingLines|None                             = None,   # Config
        seed             : IntLike|None                                       = None,   # Config
        save_path        : str|None                                           = None,   # Config
        sampler_path     : str|None                                           = None,   # Config
        data             : Data|DataList|None                                 = None,   # Data
        config           : Config|None                                        = None,   # Config
        **kwargs,
        ) -> None:
        """
        Notes
        -----
        Initialises the Acid class with inputted parameters. The class keeps calculations stored in the :py:class:`Data` class and run configurations
        in the :py:class:`Config` class (stored in Data for convenience). Both :py:class:`Data` and the :py:class:`Result` class (passed after ACID) have save and load 
        methods which can save their state, with the :py:class:`Result` class handling saving the :py:class:`Data` class together, see :ref:`result`.
        
        As of 2.0, ACID is now designed to be run on one order at a time, for running and keeping track of multiple orders, please see the :py:class:`DataList` class for a natural
        implementation of running ACID on multiple orders and keeping track of which orders have been run and which haven't, as well as storing 
        the results for each order. The :py:class:`DataList` class has been designed with parallelization on HPC's in mind, allowing orders (which are
        independent) to be run by different jobs. See also the :ref:`multiprocessing` and :ref:`datalists` sections.

        Important note: All defaults in the signature are None, meaning if any values are input, they will override the default :py:class:`Config` and/or :py:class:`Data` values or
        any values that have already been input. The defaults within the config are written below. The config defaults can also be accessed via 
        :py:attr:`ACID_code.Config.defaults` (returning a dictionary of defaults for both initialisation and the ACID method).

        All parameters below and in the ACID method are stored in the :py:class:`Config` instance, unless explicitly stated to be in the :py:class:`Data` instance.
        The :py:class:`Config` instance is for runtime settings and the :py:class:`Data` instance is for storing data and any calculations. 

        Parameters
        ----------
        velocities : :py:type:`Array1D`, optional
            Velocity grid for LSD profiles (in km/s). For example, use: np.arange(-25, 25, 0.82) to create one. If None, a default grid
            from -25 to 25 km/s is used with a spacing calculated by calc_deltav after the wavelengths are provided. It is highly recommended to 
            choose your own velocity grid, by default None, stored in the Data instance.
        linelist : :py:type:`Array2D | str` | :py:class:`LineList` | dict`, optional
            The linelist to use for LSD. The linelist should have wavelengths in angstroms and relative depths between 0 and 1.
            This is a required parameter. It can be of the forms:
            - String: A path to a VALD linelist in string format. Support for other linelists may be added in the future or on request.
            - :py:type:`Array2D`: A 2D array-like object indexed such that 0 is wavelengths and 1 is depths.
            - dict: A dictionary with keys "wavelengths" and "depths", each containing array-like objects for the wavelengths and depths respectively.
            - :py:class:`LineList`: The :py:class:`LineList` class is used to expose the linelist for masking or getting/plotting the linelist. You can input an instance if you have one.
        order : :py:type:`IntLike`, optional
            If this ACID instance is intended as a run on a specific order, then you can designate this instance for that order. This will allow
            the resulting Data instance to track of which order the profiles correspond to. Note that orders can be indexed by the correct indexing
            of the spectrograph (ie. some spectrographs start at order ~20). By default 0.
        order_range : :py:type:`Array1D`, optional
            Optionally also give ACID the full order range of the spectograph for the observation. ACID only ever runs on one order at a time,
            but this will allows ACID and eventually the DataList to keep track of which orders have been run and which haven't, and will be 
            used in the future for plotting and saving results. As with order (above), the orders can be indexed to the spectrograph orders. 
            By default [0]
        verbose : :py:type:`bool | IntLike | str`, optional
            The verbosity for printing and plotting the progress and warnings of ACID. The verbosities are natively stored as integers corresponding to:
            0: No printing or plotting, all warnings are ignored.
            1: Only printing warnings.
            2: Printing progress and warnings.
            3: Printing progress and warnings, as well as additional plots and helpful information about the run.
            The possible input types are described below:
            - Integer: Must be between 0 and 3, corresponding to the verbosities described above.
            - Boolean: If True, defaults to 2. If False, defaults to 0.
            - String: Can be one of ["none", "low", "medium", "high"] or their common variants.
            By default 2 (medium).
        sampler_progress : :py:type:`bool`, optional
            A verbosity override for just the MCMC sampling progress.
            By default None which does not override, but if True/False, it will overwrite with that value, and use/don't use a tqdm output for the sampler.
        masking_lines : :py:type:`dict` | :py:class:`MaskingLines`, optional
            Telluric lines (in angstroms) and widths in (km/s) to mask from the wavelength regions from. Unless you'd like to change the default masking
            lines, we recommend just using the defaults (leaving this as None), which are based on telluric lines and strong hydrogen/metal lines in the 
            optical and near infrared. For a guide on using your own/modifying the defaults, see :ref:`masking_lines`. By default None, stored in the Config instance.
        seed : :py:type:`IntLike`, optional
            Random seed for reproducibility, leave it on None for a random seed, by default None.
        save_path : :py:type:`str`, optional
            The path to save the data instance (containing the results) to. If None, results are not saved to disk, by default None.
            If a string is input, the data instance will be saved to this path as a .pkl file when the results are finished.
            Should be a valid file path that ends with ".pkl". If the directory containing it does not exist, it will be created.
            If a file already exists at this path, it will be overwritten on Acid initialization.
        sampler_path : :py:type:`str`, optional
            The path to save the sampler HDF5 backend file to.
            If None, the sampler is not saved and only stored in memory. By default None.
            Note that if your path points to an existing file, it will be overwritten on Acid initialization.
            If existing, we use the emcee HDF5 backend to store and load the sampler.
            Should be a valid file path that ends with ".h5". If the directory containing it does not exist, it will be created.
            Note that if you later try and save the sampler through the data class, it is converted to a HD5 backend.
        data : :py:class:`Data` | :py:class:`DataList`, optional
            An optional backend :py:class:`Data` object to use for storing data. Allows previously calculated results to be used skipped.
            If None, a new :py:class:`Data` object is created. Please note that if the :py:class:`Data` class already has a saved ACID config
            class, then any inputs to the :py:class:`Acid` initialisation and ACID method will overwrite these config values. If a 
            :py:class:`DataList` instance is inputted, the :py:class:`Data` instance corresponding to the inputted order is used.
        config : :py:class:`Config`, optional
            An optional :py:class:`Config` object to use for storing the configuration. Allows you to override the config values stored in 
            the :py:class:`Data` object, otherwise, inputs to the initialisation here and the ACID method will overwrite these config values again (if entered).
            If None, an empty Config is created and stored in the Data instance.
        **kwargs : :py:type:`dict`, optional
            Unused except to catch if users use the "linelist_path" input rather than the now "linelist" input.

        Raises
        ------
        BeartypeError
            See :ref:`type_validation` to understand input validation errors.
        """
        # Initialise the data class to store calculations in ACID        
        if data is not None:
            if isinstance(data, DataList):
                self.data = data[order]
            else:
                self.data = data
        else:
            self.data = Data() # generates also a config on initilisation

        # If a config is inputted, this will overwrite any config values already in the data class, 
        # otherwise, the config values in the data class will be used and updated by any inputs to the init or ACID method. 
        if config is not None:
            self.data.config = config

        self.config = self.data.config # Either was None (on Data initialisation above) or had a previous config stored in the old or
        # overwritten Data class

        # Validate velocities input, if None, this is handled in ACID function later when a input spectrum is provided
        if velocities is not None:
            if velocities.ndim != 1:
                raise ValueError("'velocities' must be a one-dimensional array")
        # data.velocities defaults to None in Data class, will be set in ACID function from wavelengths if not provided
        self.data.velocities = velocities if velocities is not None else self.data.velocities

        # Verbosity validation handled in config property setter
        self.config.verbose = verbose
        self.config.sampler_progress = sampler_progress

        # Catch for the linelist_path, linelist_wl, or linelist_depths arguments, which was old way to input a linelist
        if "linelist_path" in kwargs:
            linelist = kwargs.pop("linelist_path")
            if self.config.verbose > 0:
                print("Warning: 'linelist_path' is a legacy argument for inputting a linelist, " \
                f"please use 'linelist' instead.\n The 'linelist_path' argument does not support full input validation.")
        if "linelist_wl" in kwargs or "linelist_depths" in kwargs:
            raise ValueError("The 'linelist_wl' and 'linelist_depths' arguments are legacy linelist arguments, use 'linelist' instead.\n" \
                             "If your linelist wl and depths are two 1D arrays, you can use linelist=np.array([wl, depths]) for the correct format.")
        # Anything left in kwargs is invalid
        if kwargs:
            raise TypeError(
                f"Unexpected keyword argument(s) for Acid.__init__: {', '.join(sorted(kwargs))}"
            )

        # Set linelist in the Data class, the property setter handles input validation
        self.data.linelist = linelist

        # Set the lines to mask, the property in the class handles input validation and None check
        self.config.masking_lines = masking_lines

        # Set seed if not already done in config, in this way, seed is only explicitly set once
        if getattr(self.config, "seed", None) is None:
            self.config.seed = seed
            if self.config.seed is not None:
                np.random.seed(self.config.seed) # In principle this is only ever called once
            # else: user may define a seed at the top of their seed, so can use that
        # else: seed already in config, so seed would already have been set when put in

        # Set default order and order range for ACID, config handles if either of these are None
        self.config.order_range = order_range
        self.config.order = order

        # Handle data saving paths checks
        if save_path is not None:
            if not save_path.endswith(".pkl"):
                raise ValueError("'save_path' must end with '.pkl'.")
        self.config.save_path = save_path

        # Handle sampler path checks
        if sampler_path is not None:
            if not sampler_path.endswith(".h5"):
                raise ValueError("'sampler_path' must end with '.h5'.")
            # Delete existing file at sampler path if it exists
            if os.path.exists(sampler_path):
                if self.config.verbose > 0:
                    print(f"Warning: A file already exists at '{sampler_path}', it will now be deleted.")
                os.remove(sampler_path)
        self.config.sampler_path = sampler_path

        return

    # Get init keys to be checked in ACID function for any potential conflicts in input arguments.
    # This is to avoid confusion for users who may accidentally input an argument that is meant for
    # the class initialisation rather than the ACID function, which takes different arguments.
    _INIT_KEYS = set(inspect.signature(__init__).parameters) - {"self", "kwargs"}

    def ACID(
        self,
        wavelengths           : Array1D|Array2D|None        = None,   # Data
        flux                  : Array1D|Array2D|None        = None,   # Data
        errors                : Array1D|Array2D|None        = None,   # Data
        sn                    : Array1D|Array2D|Scalar|None = None,   # Data
        deterministic_profile : bool|None                   = None,   # Config
        poly_ord              : IntLike|None                = None,   # Config
        continuum_percentile  : IntLike|None                = None,   # Config
        bin_size              : IntLike|None                = None,   # Config
        pix_chunk             : IntLike|None                = None,   # Config
        dev_perc              : IntLike|None                = None,   # Config
        n_sig                 : IntLike|None                = None,   # Config
        skips                 : IntLike|None                = None,   # Config
        od                    : bool|None                   = None,   # Config
        sampler_type          : str|None                    = None,   # Config
        parallel              : bool|None                   = None,   # Config
        cores                 : IntLike|None                = None,   # Config
        nwalkers              : IntLike|None                = None,   # Config, then Data just before MCMC
        nsteps                : IntLike|None                = None,   # Config as the initial steps, Data.nsteps is the true count of steps taken, which can be higher
        max_steps             : IntLike|None                = None,   # Config
        check_interval        : IntLike|None                = None,   # Config
        min_checks            : IntLike|None                = None,   # Config
        min_tau_factor        : IntLike|None                = None,   # Config
        tau_tol               : float|None                  = None,   # Config
        moves                 : list|None                   = None,   # Config
        run_mcmc              : bool|None                   = None,   # Config
        _all_frames                                         = None,   # To work with legacy code, not to be used, silently ignored
        **kwargs,
        ) -> Result | None:
        """
        Notes
        -----
        Fits the continuum of the given spectra and performs LSD on the continuum corrected spectra,
        returning an LSD profile for each spectrum given. Spectra must cover a similiar wavelength range.

        Important note: All defaults in the signature are None, meaning if any values are input, they will override the default Config and/or Data values or
        any values that have already been input. The defaults within the config are written below. The config defaults can also be accessed by:
        ACID_code.Config.defaults (returning a dictionary of defaults for both initialisation and the ACID method).

        All parameters below are stored in the :py:class:`Config` instance, unless explicitly stated to be in the :py:class:`Data` instance.
        The :py:class:`Config` instance is for runtime settings and the :py:class:`Data` instance is for storing data and any calculations. 

        Parameters
        ----------
        wavelengths : :py:type:`Array1D | Array2D`, optional
            An array of wavelengths for each frame (in Angstroms). For multiple frames this should be a 2D array such that
            wavelengths[i] corresponds to the wavelengths for the ith frame. Can only be None if a data instance was 
            provided in initialisation. If a 2D array is provided, they are treated as multiple frames (not orders), by default None, stored in the Data instance.
        flux : :py:type:`Array1D | Array2D`, optional
            An array of spectral frames (in flux). For multiple frames this should be a 2D array such that 
            flux[i] corresponds to the spectral fluxes for the ith frame. Can only be None if a data instance was 
            provided in initialisation. If a 2D array is provided, they are treated as multiple frames (not orders), by default None, stored in the Data instance.
        errors : :py:type:`Array1D | Array2D`, optional
            Errors for each frame (in flux). For multiple frames this should be a 2D array such that
            errors[i] corresponds to the spectral errors for the ith frame. If a 2D array is provided, they are treated as multiple frames (not 
            orders). If no errors are provided, but the SN is provided, the errors will be estimated from the flux and SN, but we highly recommend 
            providing errors if possible, by default None, stored in the Data instance.
        sn : :py:type:`Scalar | IntLike | Array1D`, optional
            Average signal-to-noise ratio for each frame (used to calculate minimum line depth to consider from line list).
            Each frame should have only one S/N value, so for multiple frames this should be a 1D array such that
            sn[i] corresponds to the S/N for the ith frame. If you prefer to use a per-pixel SN value, ACID will use the :py:function:`utils.collapse_SNR` 
            function to calculate a single S/N value for each frame from the central 2/3rds of the input spectra. In which case, a 2D array can be 
            If None, the S/N will be estimated from the input spectra and errors, by default None, stored in the Data instance.
        deterministic_profile : bool, optional
            If True, fits both the continuum and the LSD profile simultaneously. If False, only fits the continuum in mcmc, the
            profile is inferred from the continuum fit. This is a new feature that has been set to the default as it significantly
            decrease convergence time and computation time per step, while fully maintaining accuracy. Setting this to False will 
            match legacy behaviour, by default True.
        poly_ord : :py:type:`IntLike`, optional
            Order of polynomial to fit as the continuum, by default 3
        continuum_percentile : :py:type:`IntLike`, optional
            The percentile to use when fitting the continuum, by default 90. For example, if 90, the continuum fit will be performed
            on the points in the spectra that are above the 90th percentile in flux in each spectral bin (determined by bin_size below).
        bin_size : :py:type:`IntLike`, optional
            The size of bins to use when performing the continuum fit. The spectra are split into bins with this number of pixels, and 
            the continuum is fit to the median wavelength and the specified percentile of flux in each bin. By default 100 pixels.
        pix_chunk : :py:type:`IntLike`, optional
            Size of 'bad' regions in pixels. 'bad' areas are identified by the residuals between an inital model
            and the data. If the residuals deviate by a specified percentage (see dev_perc below) for this number (pix_chunk) of pixels,
            then this chunk of pixels are masked in the spectra. By default 20
        dev_perc : :py:type:`IntLike`, optional
            Allowed deviation percentage. 'bad' areas are identified by the residuals between an inital model
            and the data. If a residual deviates by this percentage for a specified number of pixels,
            then this chunk of pixels are masked in the spectra. By default 25
        n_sig : :py:type:`IntLike`, optional
            Number of sigma to keep in sigma clipping. Ill fitting lines are identified by sigma-clipping the
            residuals between an inital model and the data. Regions that lie outside the median +- n_sig STDEVs are clipped.
            The clipped regions will be masked in the spectra. This masking is only applied to find the continuum fit and is removed when
            LSD is applied to obtain the final profiles, by default 3
        skips : :py:type:`IntLike`, optional
            An option to only run acid on one in every n pixels, where n is the integer argument. This is only useful for
            testing to get a quicker result especially for larger wavelength ranges or datasets, by default 1 (no skipping)
        od : :py:type:`bool`, optional
            If True, runs ACID in optical depth, otherwise, the LSD methods and ACID fitting is performed in flux. By default None which defaults to True.
            Note that the whole point of ACID is to run LSD in OD, we highly recommend leaving this unless you specifically want to compare.
        sampler_type : :py:type:`str`, optional
            If you really try to wish to use the dynesty nested sampler, you can set this to "dynesty". It is almost entirely unsupported
            by the rest of the code other than to just get a finished result object, and much slower. We highly recommend using None or "emcee" (default).
            The only reason I added this was to get the Bayesian evidence for model comparison.
            If "dynesty" is chosen, the dynesty package needs to be installed, and the nsteps parameter is treated as "nlive" to be passed to the NestedSampler.
        parallel : :py:type:`bool`, optional
            If True uses multiprocessing to calculate the profiles for each frame in parallel, see
            https://acid-code.readthedocs.io/en/stable/using_ACID.html#multiprocessing for more details. By default True
        cores : :py:type:`IntLike`, optional
            Number of cores to use if parallel=True. If None, all available cores will be used, by default None
        nwalkers : :py:type:`IntLike`, optional
            A manual override for the number of walkers for the MCMC sampler. By default, uses the emcee recommendation which is 3 times the number of
            dimensions. For the deterministic model, this is just the poly_ord + 1, for the non-deterministic model, it is poly_ord + 1 + nvelocity points.
        nsteps : :py:type:`IntLike`, optional
            Number of steps for the MCMC to run, by default 10000, the initial steps are stored in the config as nsteps, 
            but the true count of steps taken is stored in the Data instance as Data.nsteps, which can be higher than this if 
            continue_sampling is used to continue sampling after the initial run.
        max_steps : :py:type:`IntLike`, optional
            If set, the sampler will run until max_steps or convergence is reached by estimation using the emcee autocorrelation 
            time (tau). The sampler will check for convergence every 'check_interval' steps, and will require a minimum number 
            of checks ('min_checks') and a minimum tau factor ('min_tau_factor') before it can stop. The stopping criterion 
            is met when the change in tau is less than 'tau_tol' for all parameters. By default None, which means no maximum. 
            If a value is inputted, the nsteps parameter is ignored. The continue_sampling method in Result or Acid can still
            be used normally to continue sampling after either stopping criterion is reached.
        check_interval : :py:type:`IntLike`, optional
            Interval (in steps) at which to check for MCMC convergence if max_steps is set, by default 1000. 
            Only used if max_steps is set.
        min_checks : :py:type:`IntLike`, optional
            Minimum number of checks before MCMC can be stopped, by default 1. Only used if max_steps is set.
        min_tau_factor : :py:type:`IntLike`, optional
            Minimum tau factor for MCMC stopping criterion, by default 50, which is the emcee recommendation, it's not
            recommend to set a value below 50 unless you want to force convergence for the deterministic_profile=False option.
            Only used if max_steps is set.
        tau_tol : :py:type:`float`, optional
            Tolerance for tau convergence in MCMC stopping criterion, by default 0.1. Only used if max_steps is set.
        moves : list[tuple], optional
            A list of tuples specifying the moves for the MCMC sampler. The format
            tries to follow the emcee documentation as closely as possible.
            However, the config cannot store classes directly, so move names are
            used instead and converted when building the sampler.

            Each tuple should have the form::

                (move_name: str, fraction: float, move_kwargs: dict | None)

            where:

            - "move_name" is the name of the emcee move. Supported variants currently
              include "RedBlueMove", "StretchMove", "WalkMove",
              "KDEMove", "DEMove", "DESnookerMove", "MHMove",
              and "GaussianMove". Refer to the emcee documentation for more
              details on each move type. Input move names are checked against the
              "emcee.moves" module, so other moves from that module will work.
            - "fraction" is the fraction of walkers to which this move should be applied.
            - "move_kwargs" is an optional dictionary of keyword arguments passed to
              the move class initialisation.
        run_mcmc : :py:type:`bool`, optional
            If True, runs the MCMC to fit the model, by default True. Can be set to False to perform all of the preparation
            for MCMC without actually running it. The ACID function will still update the class and data attributes.
            If True, the method returns a :py:class:`Result` object, and if False, the method returns None, but attributes are updated.
        **kwargs : :py:type:`dict`, optional
            Unused except to catch accidental inputs of initialisation arguments into the ACID method and warn if so.

        Returns
        -------
        Result | None
            A :py:class:`Result` object containing the LSD profiles and associated data.
            See the :py:class:`Result` class for available methods and attributes.

            If "run_mcmc" is False, "None" is returned, but the class
            attributes are still updated.

        Raises
        ------
        BeartypeError
                See :ref:`type_validation` to understand input validation errors.
        ValueError
            If other input arguments do not conform to the expected formats and requirements.
        """
        # --- Setup and validation ---

        init_t0 = time.time()
        if self.config.verbose>1:
            print('Initialising...')

        # Check for any potential conflicts in input arguments that are meant for the class initialisation.
        overlap = self._INIT_KEYS & kwargs.keys()
        if overlap and self.config.verbose > 0:
            for key in sorted(overlap):
                print(f"'{key}' is set in Acid initialisation, not the ACID method. The inputted value will be ignored.")

        # Raise an error if the kwargs are not part of the ACID init
        invalid_keys = set(kwargs.keys()) - self._INIT_KEYS
        if invalid_keys:
            raise ValueError(
                f"The following keyword arguments are not recognised by ACID or Acid.__init__: "
                f"{', '.join(sorted(invalid_keys))}."
            )

        # Assign inputted configuration to config dictionary plus or minus a few variables
        ACID_config = {
            "poly_ord"              : poly_ord,
            "continuum_percentile"  : continuum_percentile,
            "bin_size"              : bin_size,
            "pix_chunk"             : pix_chunk,
            "dev_perc"              : dev_perc,
            "n_sig"                 : n_sig,
            "skips"                 : skips,
            "od"                    : od,
            "sampler_type"          : sampler_type,
            "parallel"              : parallel,
            "cores"                 : cores,
            "nwalkers"              : nwalkers,
            "deterministic_profile" : deterministic_profile,
            "nsteps"                : nsteps,
            "max_steps"             : max_steps,
            "check_interval"        : check_interval,
            "min_checks"            : min_checks,
            "min_tau_factor"        : min_tau_factor,
            "tau_tol"               : tau_tol,
            "moves"                 : moves,
            "run_mcmc"              : run_mcmc,
        }

        # Update config if any of the above config settings are new
        self.config.update_hipri(**ACID_config) # self.config overwrites ACID_config if overlapping
        self.data.config = self.config # update dataclass config as well, although I think this line is redundant

        if self.config.parallel and sys.platform == "win32":
            if self.config.verbose > 0:
                # This doesn't work, needs serious modifications to make work, so just run serially for now
                print("Parallel MCMC on Windows is not currently supported. Running MCMC serially.")
            self.config.parallel = False

        if self.config.sampler_type == "dynesty":
            if dynesty is None:
                raise ImportError("The 'dynesty' sampler requires the 'dynesty' package to be installed.\nPlease install it with 'pip install dynesty' or choose a different sampler type.")
        if self.config.sampler_type == "dynesty" and not self.config.deterministic_profile:
            raise ValueError("The 'dynesty' sampler can only be run with deterministic_profile=True (otherwise you'll be waiting hours for a single result)")
        if self.config.sampler_type == "dynesty" and self.config.max_steps is not None:
            raise ValueError("Cannot use max_steps as dynesty already natively supports this with live points, set nsteps=nlive. See the dynesty docs for more details.")

        # --- Start of the ACID method ---

        # Setup and data validation done in data class and applies skips
        self.data.set_inputs(wavelengths, flux, errors, sn)

        # Now that the data is set, we can check if the velocities were set in the initialisation or not, and if not,
        # calculate a default velocity grid using the input wavelengths.
        if self.data.velocities is None:
            if self.config.verbose > 0:
                print("Velocity grid not input, using a grid calculated from input wavelengths with default range of -25 to 25 km/s.\n " \
                "It is highly recommended to input your own velocity grid, especially if you need a different wavelength range.")
            deltav = utils.calc_deltav(self.data.wavelengths["input"][0])
            self.data.velocities = np.arange(-25, 25 + deltav, deltav) # default velocity grid from -25 to 25 km/s with spacing calculated from input wavelengths

        ### Begin ACID process

        # Combines spectra from each frame (weighted based of S/N), returns to S/N of combined spectra.
        # If only one frame, just uses that frame (handled in the function).
        # This function requires assigned values:
        # self.data.wavelengths["input"], self.data.flux["input"], self.data.errors["input"], self.data.sn["input"]
        # To generate:
        # As of 1.0.4, this generates self.wavelengths["combined"], self.flux["combined"], self.errors["combined"]
        # As of 1.4.0, this now instead goes to the data class, so generates self.data.wavelengths["combined"], etc.
        # As of 1.4.0, this procedure is skipped if the outputs already exists in self.data to avoid recalculation
        if all((
            hasattr(self.data.wavelengths, "combined"),
            hasattr(self.data.flux, "combined"),
            hasattr(self.data.errors, "combined"),
            hasattr(self.data.sn, "combined"),
        )):
            if self.config.verbose > 2:
                print("Combined spectra already exists, skipping combination step.")
        else:
            if self.config.verbose > 2:
                print("Combining spectra...")
            self.combine_spec(output=False)

            # Clean combined spectra of NaNs
            wavelengths, flux, errors, nanmask = utils.drop_invalid(self.data.wavelengths["combined"], self.data.flux["combined"],
                                                                    self.data.errors["combined"], return_mask=True)
            self.data.wavelengths["combined"] = wavelengths
            self.data.flux["combined"] = flux
            self.data.errors["combined"] = errors
            self.data.nanmask = nanmask

        # Perform line masking before initial fit to avoid ill-fitting lines biasing the continuum fit
        # The code for telluric masking is contained without the MaskingLines class, which both telluric_lines
        # and hydrogen_lines are instances of.
        line_mask = self.config.masking_lines.get_masks(self.data.wavelengths["combined"])
        if line_mask != []:
            line_mask = np.all(line_mask, axis=0)
            self.data.errors["combined"][line_mask] = 1e12

        # Get the initial polynomial coefficents
        if not hasattr(self.data.wavelengths, "combined_normalized"):
            a, b = utils.get_normalisation_coeffs(self.data.wavelengths["combined"])
            self.data.wavelengths["combined_normalized"] = (self.data.wavelengths["combined"]*a)+b

        # Compute an initial continuum fit
        # poly inputs has polynomial coefficients and scale at the end
        if all((
            hasattr(self.data.flux, "fitted"),
            hasattr(self.data.errors, "fitted"),
            self.data.poly_inputs is not None
        )):
            if self.config.verbose > 2:
                print("Continuum fit already exists, skipping initial fit step.")
        else:
            if self.config.verbose > 2:
                print("Performing initial continuum fit...")
            self.data.poly_inputs, self.data.flux["fitted"], self.data.errors["fitted"] = self.continuumfit(
                self.data.flux["combined"],
                self.data.wavelengths["combined"],
                self.data.errors["combined"],
                poly_ord = self.config.poly_ord,
                plot_result = self.config.verbose > 2,
                plot_type = "initial"
            )
        self.data.wavelengths["fitted"] = np.copy(self.data.wavelengths["combined"]) # Just to keep track
        self.data.sn["fitted"]          = np.copy(self.data.sn["combined"]) # SN also is not changed here

        # Get the initial LSD profile and set the alpha matrix (unchanged from masking) and model_inputs
        if all((
            self.data.model_inputs is not None,
            self.data.alpha is not None
        )):
            if self.config.verbose > 1:
                print("Initial LSD profile already exists, skipping initial LSD step.")
        else:
            if self.config.verbose > 1:
                print("Calculating initial LSD profile...")
            # Get the initial LSD profile using the initial fit
            initial_LSD = LSD(self.data) # Initialise LSD class with standard Acid attributes (verbosity, linelist, velocities, etc)
            initial_LSD.run_LSD(self.data.wavelengths["fitted"], self.data.flux["fitted"], self.data.errors["fitted"], self.data.sn["fitted"])

            # Use alpha matrix and initial profile class variables from initial LSD run
            self.data.initial_profile = initial_LSD.profile # in optical depth
            self.data.initial_profile_errors = initial_LSD.profile_errors # Not used, saved for debugging
            self.data.alpha = initial_LSD.alpha
            # Set x, y, yerr, and model_inputs for emcee
            self.data.model_inputs = np.concatenate((self.data.initial_profile, self.data.poly_inputs))

        # Masking based off residuals
        # Requires: self.x, self.y, self.yerr, self.data.model_inputs, self.poly
        # Sets: self.c_factor, self.residual_masks
        # Modifies: self.yerr, and as of 1.5, self.data.model_inputs
        # As of 1.4.0, this is all applied to the data class (not internal ACID variables)
        if all((
            self.data.residual_masks is not None,
            self.data.c_factor is not None
        )):
            if self.config.verbose > 1:
                print("Residual masks already exists, skipping residual masking step.")
        else:
            if self.config.verbose>1:
                print('Residual masking...')
            self.residual_mask()

        # Setting number of walkers and their start values(pos)
        self.data.ndim = len(self.data.model_inputs) if not self.config.deterministic_profile else self.config.poly_ord + 1
        # emcee recommendation is 3 times the number of dimensions, but can be overridden by user input
        self.data.nwalkers = self.data.ndim * 3 if self.config.nwalkers is None else self.config.nwalkers
        rng = np.random.default_rng(self.config.seed)

        # Starting values of walkers with independent variation#
        if self.config.sampler_type == "emcee":
            sigma = 0.8 * 0.005
            initial_state = []
            for i in range(0, len(self.data.model_inputs)):
                if i < len(self.data.velocities):
                    if not self.config.deterministic_profile:
                        pos = rng.normal(self.data.model_inputs[i], sigma, (self.data.nwalkers, ))
                    else:
                        continue
                else:
                    x1 = self.data.model_inputs[i]
                    rounded_sigma = round(x1, 1-int(floor(log10(abs(x1))))-1)
                    sigma = abs(rounded_sigma) / 10
                    pos = rng.normal(self.data.model_inputs[i], sigma, (self.data.nwalkers, ))
                initial_state.append(pos)
            initial_state = np.array(initial_state).T
        else:
            initial_state = None

        ### ACID initialialised ###
        self.data.setup_time += time.time() - init_t0
        mcmc_t0 = time.time()
        if self.config.verbose>1:
            print('Initialised in %ss'%round((self.data.setup_time), 3))
        if self.config.verbose>2:
            print('ACID Configuration before MCMC run:')
            print(f"Polynomial order: {self.config.poly_ord}")
            print(f"Deterministic profile: {self.config.deterministic_profile}")
            print(f"Number of walkers: {self.data.nwalkers}")
            print(f"Number of dimensions: {self.data.ndim}")

        # Run MCMC
        if self.config.run_mcmc is True:
            # Default run for just nsteps steps
            if self.config.max_steps is None:
                if self.config.verbose > 1:
                    print("Running MCMC for %s steps..."%self.config.nsteps)
                self.run_mcmc(self.config.nsteps, initial_state)
                self.data.nsteps += self.config.nsteps
            # Else use max_steps path
            else:
                if self.config.verbose > 1:
                    print(f"Running MCMC with a maximum of {self.config.max_steps} steps or until convergence is reached...")
                self.run_mcmc_until_converged(self.config.max_steps, initial_state)
                self.data.nsteps = self.step_number
            self.data.mcmc_time += time.time() - mcmc_t0

            if self.config.verbose>1:
                print('MCMC finished after %ss'%(round(self.data.mcmc_time, 3)))

            return Result(self)

        else:
            if self.config.verbose > 0:
                print("MCMC not run, returning None. Class attributes have been updated.")
            return None

    def ACID_HARPS(self, *args, **kwargs):
        """
        This method is no longer supported in ACID. Please use the ACID function with the appropriate inputs for HARPS spectra instead. 
        Future versions of ACID will provide functions to load and configure data from a range of different standard instruments. 
        If you still really wish to use ACID_HARPS, the last stable version of ACID with the method is 1.4.5. Try: pip install ACID_code_v2==1.4.5
        """
        raise NotImplementedError(f"ACID_HARPS is no longer supported in ACID. \n"
        f"Please use the ACID function with the appropriate inputs for HARPS spectra instead. \n"
        f"Future versions of ACID will provide functions to load and configure data from a range of different standard instruments. \n"
        f"If you still really wish to use ACID_HARPS, the last stable version of ACID with the method is 1.4.5. Try: pip install ACID_code_v2==1.4.5")

    def combine_spec(
        self,
        frame_wavelengths: Array1D|Array2D|None = None,
        frame_flux:        Array1D|Array2D|None = None,
        frame_errors:      Array1D|Array2D|None = None,
        frame_sns:         Array1D|Array2D|None = None,
        output:            bool                 = True
        ) -> tuple | None:
        """
        Combines the multiple inputted spectral frames into one spectrum, or just passes through the single frame if only one was input. 
        The frames are interpolated onto a common wavelength grid of the spectrum with the highest S/N, and then a weighted average is used based on the errors. 
        The S/N of the combined spectrum is also calculated based on the input S/N and the weights.

        Parameters
        ----------
        frame_wavelengths : :py:type:`Array1D` | :py:type:`Array2D`, optional
            Wavelengths for the spectral frames, by default None
        frame_flux : :py:type:`Array1D` | :py:type:`Array2D`, optional
            Fluxes for the spectral frames, by default None
        frame_errors : :py:type:`Array1D` | :py:type:`Array2D`, optional
            Errors for the spectral frames, by default None
        frame_sns : :py:type:`Array1D` | :py:type:`Array2D`, optional
            Signal-to-noise ratio for the spectral frames, by default None
        output : bool, optional
            Whether to output the combined spectrum, by default True

        Returns
        -------
        tuple | None, if output is True, containing:
            combined_wavelengths : np.ndarray
                Wavelengths for the combined spectrum
            combined_spectrum : np.ndarray
                Fluxes for the combined spectrum
            combined_errors : np.ndarray
                Errors for the combined spectrum
            combined_sn : float
                Signal-to-noise ratio for the combined spectrum
        None, if output is False, but the combined spectrum is still saved in the data class attributes.
        """

        if frame_wavelengths is not None: # This should only be for testing
            self.data.wavelengths["input"] = np.copy(frame_wavelengths)
            self.data.flux["input"]        = np.copy(frame_flux)
            self.data.errors["input"]      = np.copy(frame_errors)
            self.data.sn["input"]          = np.copy(frame_sns)

        # Set simple names for variables (just used in this function)
        wavelengths = np.copy(self.data.wavelengths["input"])
        flux        = np.copy(self.data.flux["input"])
        errors      = np.copy(self.data.errors["input"])
        sn          = np.copy(self.data.sn["input"])

        # Return as is if only one spectrum
        if len(self.data.wavelengths["input"])==1:
            self.data.wavelengths["combined"] = np.copy(self.data.wavelengths["input"][0])
            self.data.flux["combined"]        = np.copy(self.data.flux["input"][0])
            self.data.errors["combined"]      = np.copy(self.data.errors["input"][0])
            self.data.sn["combined"]          = np.copy(self.data.sn["input"][0])

        else:
            # Get wavelength grid with highest S/N
            combined_wavelengths = wavelengths[np.argmax(sn)]

            interpolated_flux   = np.zeros_like(flux)
            interpolated_errors = np.zeros_like(errors)

            # combine all spectra to one spectrum
            for n in range(len(flux)):

                # Interpolate each spectrum onto the combined wavelength grid
                f2 = interp1d(wavelengths[n], flux[n], kind = 'linear', bounds_error=False, fill_value = 'extrapolate')
                f2_err = interp1d(wavelengths[n], errors[n], kind = 'linear', bounds_error=False, fill_value = 'extrapolate')
                interpolated_flux[n] = f2(combined_wavelengths)
                interpolated_errors[n] = f2_err(combined_wavelengths)

                # Mask out out extrapolated areas
                idx_ex = np.logical_and(combined_wavelengths<=np.max(wavelengths[n]),
                                        combined_wavelengths>=np.min(wavelengths[n]))
                idx_ex = tuple([idx_ex==False])

                interpolated_flux[n][idx_ex] = 1.
                interpolated_errors[n][idx_ex] = 1e12

                # Mask out nans and zeros (these do not contribute to the main spectrum)
                where_are_NaNs = np.isnan(interpolated_flux[n])
                interpolated_errors[n][where_are_NaNs] = 1e12
                where_are_zeros = np.where(interpolated_flux[n] == 0)[0]
                interpolated_errors[n][where_are_zeros] = 1e12

                where_are_NaNs = np.isnan(interpolated_errors[n])
                interpolated_errors[n][where_are_NaNs] = 1e12
                where_are_zeros = np.where(interpolated_errors[n] == 0)[0]
                interpolated_errors[n][where_are_zeros] = 1e12

            invvars = 1 / interpolated_errors**2
            invvars[interpolated_errors >= 1e12] = 0

            weights = np.sum(invvars, axis=0)
            non_zero = weights > 0
            
            weighted_flux   = np.sum(interpolated_flux * invvars, axis=0)
            
            combined_flux = np.full_like(weights, 1.0)      # or np.nan
            combined_errors = np.full_like(weights, 1e12)

            combined_flux[non_zero] = weighted_flux[non_zero] / weights[non_zero]
            combined_errors[non_zero] = 1 / np.sqrt(weights[non_zero])

            frame_weights = np.sum(invvars, axis=1)
            combined_sn   = np.sum(frame_weights * sn) / np.sum(frame_weights)

            self.data.wavelengths["combined"] = combined_wavelengths
            self.data.flux["combined"]        = combined_flux
            self.data.errors["combined"]      = combined_errors
            self.data.sn["combined"]          = combined_sn

        if output is True:
            # ie if called as a function rather than from ACID function
            return combined_wavelengths, combined_flux, combined_errors, combined_sn

    def continuumfit(
        self,
        fluxes      : Array1D,
        wavelengths : Array1D,
        errors      : Array1D,
        poly_ord    : IntLike = 3,
        plot_result : bool    = False,
        plot_type   : str     = "initial"
        ) -> tuple:
        """Provides an initial, normalised continuum fit using inputted spectra.

        Parameters
        ----------
        fluxes : np.ndarray
            The flux values of the spectrum.
        wavelengths : np.ndarray
            The wavelengths corresponding to the spectrum.
        errors : np.ndarray
            The error values associated with the spectrum.
        poly_ord : int
            The order of the polynomial to fit to the continuum. By default 3.
        plot_result : bool, optional
            Whether to plot the continuum fit result, by default False.
        plot_type : str, optional
            The type of plot to generate, either "initial" or "masked", by default "initial"

        Returns
        -------
        tuple containing:
            - Polynomial coefficients: The coefficients of the fitted polynomial, ordered from highest degree to lowest.
            - Normalized flux: The flux values normalized by the fitted continuum.
            - Normalized errors: The error values normalized by the fitted continuum.
        """
        # Normalise wavelengths
        a, b = utils.get_normalisation_coeffs(wavelengths)
        unnormalized_wavelengths = np.copy(wavelengths)
        wavelengths = (wavelengths*a)+b

        # Sort to ensure smooth binning and fitting
        idx = np.argsort(wavelengths)
        w = wavelengths[idx]
        f = fluxes[idx]
        e = errors[idx]

        # Get nbins and bin_size, reshape into 2D array of bins
        binsize = self.config.bin_size
        n = len(w) // binsize  # full bins only
        w2 = w[:n*binsize].reshape(n, binsize)
        f2 = f[:n*binsize].reshape(n, binsize)
        e2 = e[:n*binsize].reshape(n, binsize)

        # Get the median wavelength, specified percentile flux, and median error in each bin
        clipped_flux = np.nanpercentile(f2, self.config.continuum_percentile, axis=1)
        clipped_waves = np.nanmedian(w2, axis=1)
        clipped_errs = np.nanmedian(e2, axis=1)

        # Remove bad points for the polynomial fit, defined as non-finite values or errors that are non-positive or above 1e11
        good = (
            np.isfinite(clipped_waves)
            & np.isfinite(clipped_flux)
            & np.isfinite(clipped_errs)
            & (clipped_errs > 0)
            & (clipped_errs < 1e11) # 1e12 is the default mask error value, which can be picked up in the median error binning
        )

        # Fit with np.polyfit
        coeffs = np.polyfit(clipped_waves[good], clipped_flux[good], poly_ord, w=1/clipped_errs[good])
        poly = np.poly1d(coeffs)
        fit = poly(wavelengths)

        # Get the model fitted flux and errors from the fit
        flux_obs = fluxes / fit
        new_errors = errors / fit

        # Flip coefficients for use in MCMC (ordered from lowest degree to highest)
        poly_coeffs = np.flip(coeffs)

        # Save to Data the required variables for the plot
        if plot_type not in self.data.plotting_variables:
            self.data.plotting_variables[plot_type] = {}
        self.data.plotting_variables[plot_type]["unnormalized_wavelengths"] = unnormalized_wavelengths
        self.data.plotting_variables[plot_type]["fluxes"]                   = fluxes
        self.data.plotting_variables[plot_type]["fit"]                      = fit
        self.data.plotting_variables[plot_type]["clipped_waves"]            = clipped_waves
        self.data.plotting_variables[plot_type]["clipped_flux"]             = clipped_flux
        self.data.plotting_variables[plot_type]["good"]                     = good
        if plot_result is True:
            self.data.plot_continuum_fit(plot_type=plot_type)

        if np.any(flux_obs <= 0) or np.any(new_errors <= 0):
            error = ContinuumError("Continuum fit resulted in non-positive flux or errors, which is not physical.\n " \
            "Consider adjusting the polynomial order or continuum percentile. Use verbose=3 to see the plot of the continuum fit.\n " \
            "Note that this will only work for interactive terminals or displays which work with plt.show()")
            self.data.exception = error
            self.data.traceback = traceback.format_stack()
            raise error

        return poly_coeffs, flux_obs, new_errors

    def residual_mask(self) -> None:
        """
        Masks regions of the spectrum based on residuals from an initial model fit. A purely class method not to be used elsewhere.
        This function is really only supposed to be used in the class, so no inputs are accepted. It is only used once in ACID 
        and could be put directly in the method, but this allows a clearer checkpoint which segments saving the result of the mask for analysis.
        """

        # Residual masking - mask continuous areas first - then possibly progress to masking the narrow lines

        # Set standard variables
        x = self.data.wavelengths["combined"]
        y = self.data.flux["combined"]
        yerr = self.data.errors["combined"]
        sn = self.data.sn["combined"]

        # Use the initial LSD run to get the forward model and scaled residuals
        forward, _profile = mcmc.MCMC(x, y, yerr, self.data.alpha, od=self.config.od).full_model(self.data.model_inputs)
        residuals = (y - forward) / forward

        # Chunk masking based on deviation from residuals
        # -----------------------------------------------

        # Get bad pixels that deviate by a percentage greater than dev_perc
        bad_idx = np.abs(residuals) > (self.config.dev_perc / 100)

        # A trick to get the mask for continous regions of bad pixels, by padding the bad_idx 
        # with False on both sides and finding the start and end indices of the True regions
        padded = np.concatenate(([False], bad_idx, [False]))
        starts = np.flatnonzero(~padded[:-1] & padded[1:])
        ends = np.flatnonzero(padded[:-1] & ~padded[1:])
        pix_mask = np.zeros_like(residuals, dtype=bool)

        # Then mask those regions that are greater than pix_chunk in length
        for start, end in zip(starts, ends):
            if (end - start) >= self.config.pix_chunk:
                yerr[start:end] = 1e12
                pix_mask[start:end] = True

        # Warn if more than 50% of spectrum is masked this way
        if np.sum(pix_mask) > 0.5 * len(pix_mask):
            if self.config.verbose > 0:
                print(f"Warning: More than 50% of the spectrum is masked based on residuals. \n" \
                "Please check your initial continuum fit (by using verbose=3 when initialising), \n" \
                "or consider adjusting the pix_chunk and dev_perc parameters. If you are aware that you \n" \
                "have bad spectra, then this can be ignored.")

        # Note that this is used to keep track of the residual masks for later use when processing the results
        self.data.residual_masks = tuple([yerr >= 1e12])

        # Sigma clipping
        # --------------

        # Get median, sigma, and clip limits
        m = np.median(residuals)
        sigma = np.std(residuals)
        clip = self.config.n_sig * sigma
        lower_clip = m - clip
        upper_clip = m + clip

        # Find and apply mask
        mask = (residuals <= lower_clip) | (residuals >= upper_clip)
        yerr[mask] = 1e12

        # Now do another continuum fit with masked yerr, continuumfit removes high error points from the fit
        poly_inputs, fitted_flux, fitted_errors  = self.continuumfit(y, x, yerr, self.config.poly_ord,
                                                   plot_result=self.config.verbose > 2,
                                                   plot_type="masked")

        # Run LSD again with the new fitted flux and errors
        LSD_masking = LSD(self.data)
        # Since the above ONLY modifies yerr, and the alpha matrix is independent of yerr, we can input previous 
        # alpha since it wil be the same. We still run LSD to get c_factor and the profile
        # alpha is only dependent on wavelengths and linelist, which are unchanged
        LSD_masking.run_LSD(x, fitted_flux, fitted_errors, sn, alpha=self.data.alpha)

        # Update and set new variables
        self.data.c_factor = LSD_masking.c_factor
        self.data.initial_model_inputs = np.copy(self.data.model_inputs) # Save the initial model inputs before masking for later use if needed
        self.data.model_inputs = np.concatenate((LSD_masking.profile, poly_inputs))

        # Set masked variables
        self.data.wavelengths["masked"] = x
        self.data.flux["masked"]        = y # x and y dont change in this func
        self.data.errors["masked"]      = yerr # yerr is modified in this func
        self.data.sn["masked"]          = np.copy(self.data.sn["combined"]) # SN is not changed in this func
        # self.alpha is also modified in this func to get new alpha with masked residuals using pix chunk and dev perc

        # Set required variables for plotting in the Data class
        if "residual_masking" not in self.data.plotting_variables:
            self.data.plotting_variables["residual_masking"] = {}
        self.data.plotting_variables["residual_masking"]["mask"] = mask
        self.data.plotting_variables["residual_masking"]["residuals"] = residuals
        self.data.plotting_variables["residual_masking"]["upper_clip"] = upper_clip
        self.data.plotting_variables["residual_masking"]["lower_clip"] = lower_clip
        self.data.plotting_variables["residual_masking"]["pix_mask"] = pix_mask
        self.data.plotting_variables["residual_masking"]["profile_F"] = LSD_masking.profile_F
        if self.config.verbose > 2:
            self.data.plot_residual_masking()

        return

    def run_mcmc(
        self,
        nsteps:IntLike,
        state = None,        
        ) -> None:
        """
        Runs MCMC for a specified number of steps. A purely class method that I do not recommend you use directly. Use
        Acid.ACID(run_mcmc=True) to run MCMC for the first pass if not already done, which will skip already performed calculations.
        Otherwise, use Acid.continue_sampling or Result.continue_sampling if you have already run MCMC and want to continue.
        """

        # Get default sampler kwargs from initial state
        if self.config.sampler_type == "emcee":
            sampler_kwargs, mcmc_kwargs = self._get_sampler_kwargs(nsteps, state)
        pool_context = nullcontext(None)

        if self.config.parallel:
            utils.configure_mp_environ(os) # Raises error is not configured correctly, otherwise does nothing

            if self.config.verbose>1:
                print(f"Using {self.config.cores} cores for MCMC")
            
            ctx = mp.get_context("fork")
            pool_context = ctx.Pool(processes=self.config.cores, initializer=mcmc._mp_init_worker, initargs=(self.data,))
            log_prob = mcmc._mp_log_probability if self.config.sampler_type == "emcee" else mcmc._mp_log_likelihood
            ptform = mcmc._mp_ptform
            queue_size = os.cpu_count()
        else:
            MCMC = mcmc.MCMC(self.data)
            log_prob = MCMC if self.config.sampler_type == "emcee" else MCMC.dynesty_logprob
            ptform = MCMC.ptform
            queue_size = None

        with pool_context as pool:
            if self.config.sampler_type == "emcee":
                self.sampler = EnsembleSampler(log_prob_fn=log_prob, pool=pool, **sampler_kwargs)
                self.sampler.run_mcmc(**mcmc_kwargs)
            else:
                import dynesty
                if self.config.parallel:
                    pool.size = self.config.cores
                self.sampler = dynesty.NestedSampler(log_prob, ptform, self.data.ndim, self.config.nsteps, pool=pool, queue_size=queue_size)
                self.sampler.run_nested(print_progress=self.config.verbose>1)

    def run_mcmc_until_converged(self, max_steps:IntLike, state=None) -> None:
        """
        Runs MCMC until convergence is reached. A purely class method that I do not recommend you use directly. Use
        Acid.ACID(run_mcmc=True) to run MCMC for the first pass if not already done, which will skip already performed calculations.
        Otherwise, use Acid.continue_sampling or Result.continue_sampling if you have already run MCMC and want to continue.
        """
        # Get sampler and stopping criterion kwargs for the first run based on initial state, then update nsteps in mcmc_kwargs for subsequent runs
        sampler_kwargs, mcmc_kwargs = self._get_sampler_kwargs(nsteps=self.config.check_interval, state=state)
        stopping_criterion_args = (self.config.min_checks, self.config.min_tau_factor, self.config.tau_tol)

        # Set variables to be updated within the convergence loop
        step_number = 0
        tau_list = []
        max_samples = max_steps // self.config.check_interval
        last_tolerance = np.inf
        last_neff = 0
        condition = False
        pool_context = nullcontext(None)

        if self.config.parallel:
            utils.configure_mp_environ(os)

            if self.config.verbose > 1:
                print(f"Using {self.config.cores} cores for MCMC")

            ctx = mp.get_context("fork")
            pool_context = ctx.Pool(processes=self.config.cores, initializer=mcmc._mp_init_worker, initargs=(self.data,))
            log_prob_fn = mcmc._mp_log_probability
        else:
            log_prob_fn = mcmc.MCMC(self.data)

        with pool_context as pool:
            self.sampler = EnsembleSampler(**sampler_kwargs, pool=pool, log_prob_fn=log_prob_fn)

            for i in range(max_samples):
                tol_str, neff_str = mcmc.MCMC._get_tqdm_desc(last_tolerance, last_neff, self.config)
                mcmc_kwargs["progress_kwargs"] = {"desc": f"Iteration {i+1}/{max_samples}, last tolerance: {tol_str}, neff: {neff_str}"}
                self.sampler.run_mcmc(**mcmc_kwargs, skip_initial_state_check=True)
                mcmc_kwargs["initial_state"] = None
                step_number += self.config.check_interval

                # We want to keep the time for get_autocorr_time to run constant, so thin accordingly
                # It scales with the number of steps, so thin by the number of steps taken divided by 
                # the check interval to keep the same number of samples for get_autocorr_time to process.
                try:
                    with open(os.devnull, "w") as devnull, \
                        contextlib.redirect_stdout(devnull), \
                        contextlib.redirect_stderr(devnull):
                        tau = self.sampler.get_autocorr_time(tol=0, thin=step_number//self.config.check_interval)
                except emcee.autocorr.AutocorrError:
                    continue
                tau_list.append(tau)

                # The stopping criterion function below handles the logic for determining stopping condition
                condition, last_tolerance, last_neff = mcmc.MCMC._get_mcmc_stopping_criterion(tau_list, step_number, *stopping_criterion_args)

                if condition is True:
                    if self.config.verbose > 1:
                        print(f"Converged at step {step_number}. Final tolerance: {last_tolerance:.4f}, final effective sample size: {last_neff:.2f}.")
                    break

        # Warn if convergence not reached after either parallel or non-parallel version
        if self.config.verbose > 1 and condition is False:
            print(f"Not converged after reaching max steps of {step_number}. Final effective sample size: {last_neff:.2f}, final tolerance: {last_tolerance:.4f}.\n"
                  f"Consider increasing max_steps.")

        self.step_number = step_number # Update step number once mcmc has finished
        return

    def _get_sampler_kwargs(self, nsteps, state=None):
        # Gets sampler kwargs for the emcee EnsembleSampler and run_mcmc functions based on the current state of the
        # ACID instance and the inputted nsteps and state.

        # Set verbosity of the sampler with sampler_progress override if specified
        sampler_verbosity = True if self.config.verbose>1 else False
        sampler_verbosity = self.config.sampler_progress if self.config.sampler_progress is not None else sampler_verbosity
        
        backend = None
        if state is None:
            if self.sampler is None:
                raise ValueError(f"Either a state or an existing sampler must be provided to initiate the sampler. \n" \
                                 "This has most likely happened because you ran continue_sampling without first running ACID or using run_mcmc=False.")
            backend = self.sampler.backend # This includes previous seed
        
        # Now that the backend has been set depending on an existing state, if the backend is stil None, we choose depending on sampler_path
        if backend is None and self.config.sampler_path is not None:
            os.makedirs(os.path.dirname(self.config.sampler_path), exist_ok=True)
            backend = emcee.backends.HDFBackend(self.config.sampler_path)
            if self.config.verbose > 1:
                print(f"Using sampler backend at {self.config.sampler_path}")
        # else: leave none and a normal in-memory sampler backend is used

        if self.config.cores is None:
            if "SLURM_JOB_ID" in os.environ:
                self.config.cores = int(os.environ.get("SLURM_CPUS_ON_NODE", 1))
            else:
                self.config.cores = os.cpu_count()

        # Configure moves based on config, this function converts a Config moves dictionary format to a class format
        # accepted for emcee moves.
        moves = utils.convert_moves_to_emcee(self.config.moves)

        sampler_kwargs = {
            "nwalkers"   : self.data.nwalkers,
            "ndim"       : self.data.ndim,
            "moves"      : moves,
            "backend"    : backend,
        }
        mcmc_kwargs = {
            "initial_state": state,
            "nsteps"       : nsteps,
            "progress"     : sampler_verbosity,
            "store"        : True,
            "tune"         : True
        }
        return sampler_kwargs, mcmc_kwargs

    def continue_sampling(
        self,
        nsteps           : IntLike|None = None,
        max_steps        : IntLike|None = None,
        max_steps_kwargs : dict|None    = None,
        parallel         : bool         = None,
        cores            : int          = None,
        moves            : dict         = None,
        return_sampler   : bool         = False,
        ) -> EnsembleSampler | None:
        """
        Continue MCMC sampling for additional steps. This should be called in Result class by the user.
        This necessarily requires a Data instance to have been put into the ACID init.

        Parameters
        ----------
        nsteps : :py:type:`IntLike`, optional
            Number of additional steps to run the MCMC for.
        max_steps : :py:type:`IntLike`, optional
            Maximum number of steps to run the MCMC for in total (including previous steps).
            If specified, the MCMC will stop if this number of steps is reached even if convergence has not been reached, by default None.
            If input, nsteps is ignored.
        max_steps_kwargs : dict, optional
            Additional keyword arguments to be passed to the run_mcmc_until_converged function if max_steps is specified, by default None.
            The kwargs description can be found in Acid.ACID(), they are the 4 kwargs appearing after max_steps. Typos for kwargs are silently
            ignored.
        parallel : bool, optional
            Overwrites config with whether to run the MCMC in parallel. If None, uses already existing configuration. Default is None.
        cores : int, optional
            Overwrites config with the number of cores to use for parallel MCMC. If None, uses already existing configuration. Default is None.
        moves : dict, optional
            Overwrites config with the dictionary specifying the moves to use for MCMC sampling. If None, uses already existing configuration. 
            Default is None. See :py:function:`Acid.ACID` for format.
        return_sampler : bool, optional
            Whether to return the sampler after continuing sampling. Default is True.

        Returns
        -------
        emcee.EnsembleSampler | None
            The MCMC sampler after running for the additional steps, or None if return_sampler is False.
        """
        # Update config with any new parallel, cores, or moves settings for the continued sampling
        self.config.update_hipri(parallel=parallel, cores=cores, moves=moves)

        if max_steps is not None:
            if max_steps_kwargs is not None:
                self.config.update_hipri(**max_steps_kwargs)
            self.run_mcmc_until_converged(max_steps, state=None) # continue from current state
            self.data.nsteps += self.step_number
        else:
            self.run_mcmc(nsteps, state=None) # continue from current state
            self.data.nsteps += nsteps

        if return_sampler:
            return self.sampler

    @property
    def result(self) -> Result:
        """Return a Result object for this instance or one passed explicitly.

        Returns
        -------
        Result
            The Result object for the given Acid instance.
        """
        if not self.data.complete:
            raise ValueError("ACID has not been run yet. Cannot create a Result instance.")
        return Result(self)

    @property
    def sampler(self):
        """Returns the sampler stored in the Data class."""
        return self.data.sampler

    @sampler.setter
    def sampler(self, value):
        self.data.sampler = value

# All code below is just to ensure backward compatibility with previous ACID versions
def ACID(*args, **kwargs):
    """Legacy ACID function

    This function runs the legacy ACID code. This is provided for backwards compatibility with previous versions of ACID.
    It is recommended to use the ACID class and its methods for new code. The args and kwargs passing follows the original
    pre-2.0 version of ACID, which can be found in the earlier releases in https://github.com/ldolan05/ACID

    Parameters
    ----------
    *args
        Positional arguments to be passed to the ACID function.
    **kwargs
        Keyword arguments to be passed to the ACID initialisation and function.

    Returns
    -------
    Any
        Returns the outputs of the ACID function (now a Result object).
    """
    # Use old argument names and map to new ones
    LEGACY_ACID_ARGS = [
        "input_wavelengths",
        "input_spectra",
        "input_spectral_errors",
        "line",
        "frame_sns",
        "vgrid",
        "all_frames",
        "poly_or",
        "pix_chunk",
        "dev_perc",
        "n_sig",
        "telluric_lines",
        "order",
    ]
    RENAMED_LEGACY_ARGS = {
        "input_wavelengths": "wavelengths",
        "input_spectra": "flux",
        "input_spectral_errors": "errors",
        "frame_sns": "sn",
        "vgrid": "velocities",
        "line": "linelist",
        "poly_or": "poly_ord",
        "all_frames": "_all_frames",
    }

    # Split args and kwargs into init and run kwargs using helper function
    init_kwargs, run_kwargs = _get_init_and_run_kwargs(LEGACY_ACID_ARGS, RENAMED_LEGACY_ARGS, *args, **kwargs)

    acid = Acid(**init_kwargs)
    return acid.ACID(**run_kwargs)

def ACID_HARPS(*args, **kwargs):
    """Legacy ACID_HARPS function, deprecated after 1.4.5.
    """
    raise NotImplementedError(f"ACID_HARPS is no longer supported. \n"
        f"Please use the ACID function with the appropriate inputs for HARPS spectra instead. \n"
        f"Future versions of ACID will provide functions to load and configure data from a range of different standard instruments. \n"
        f"If you still really wish to use ACID_HARPS, the last stable version of ACID with the method is 1.4.5. Try: pip install ACID_code==1.4.5")

def _get_init_and_run_kwargs(legacy_args, renamed_args_map, *args, **kwargs):
    """Helper function to split legacy args and kwargs into init and run kwargs given
    legacy argument names and their renamed counterparts.
    """
    legacy_kwargs = {}

    # Check for too many positional arguments
    if len(args) > len(legacy_args):
        raise TypeError(f"Too many positional arguments: {len(args)}")

    # Map positional arguments to their legacy names
    for i, val in enumerate(args):
        legacy_kwargs[legacy_args[i]] = val
    
    # Map legacy argument names to new ones
    translated_legacy = {}
    for key, val in legacy_kwargs.items():
        new_key = renamed_args_map.get(key, key)
        translated_legacy[new_key] = val
    translated_kwargs = {}
    for key, val in kwargs.items():
        new_key = renamed_args_map.get(key, key)
        translated_kwargs[new_key] = val

    # Combine both translated dictionaries
    combined = {**translated_legacy, **translated_kwargs}

    # Determine which arguments are for __init__ and which are for run_ACID_HARPS
    init_params = inspect.signature(Acid.__init__).parameters
    init_keys = set(init_params.keys()) - {"self"}

    # Split kwargs accordingly
    init_kwargs = {key: val for key, val in combined.items() if key in init_keys}
    run_kwargs = {key: val for key, val in combined.items() if key not in init_keys}
    return init_kwargs, run_kwargs


# Old LSD alpha matrix code, iterative step:
# The below I think works similarly to above, but but uses only for loops, rather then using a for loop
# that only involves numpy calculations. With testing, the above is much faster.
else:
    warnings.warn('Large wavelength ranges give large computation time. Seperate wavelength range into smaller chunks for faster computation.', DeprecationWarning, stacklevel=2)
    alpha = np.zeros((len(blankwaves), len(velocities)))

    for j in tqdm(range(0, len(blankwaves)), desc='Calculating alpha matrix'):
        for i in (range(0, len(self.wavelengths_expected))):
            vdiff = ((blankwaves[j] - self.wavelengths_expected[i]) * ckms) / self.wavelengths_expected[i]
            if vdiff <= (np.max(velocities) + deltav) and vdiff >= (np.min(velocities) - deltav):
                diff = blankwaves[j] - self.wavelengths_expected[i]
                vel = const.c / 1e3 * (diff / self.wavelengths_expected[i])
                for k in range(0, len(velocities)):
                    x = (velocities[k] - vel) / deltav
                    if -1. < x and x < 0.:
                        delta_x = (1 + x)
                        alpha[j, k] = alpha[j, k] + depths_expected[i] * delta_x
                    elif 0. <= x and x < 1.:
                        delta_x = (1 - x)
                        alpha[j, k] = alpha[j, k] + depths_expected[i] * delta_x
            else:
                pass

# Vectorized old part:
# The below was simplified using np.clip (as shown above)
            alpha_mask_1 = np.logical_and(-1. < x, x < 0.)
            alpha_mask_2 = np.logical_and(0. <= x, x < 1.)
            delta_x_1 = 1 + x
            delta_x_2 = 1 - x
            delta_x_1[alpha_mask_1==False] = 0
            delta_x_2[alpha_mask_2==False] = 0
            # Update alpha array using calculated delta_x values
            alpha = np.zeros((len(blankwaves), len(velocities)))
            alpha += (depths_expected[:, np.newaxis] * delta_x_1).sum(axis=1)
            alpha += (depths_expected[:, np.newaxis] * delta_x_2).sum(axis=1)

# Old blaze correct code part 1:
# ## test - s1d interpolated onto e2ds wavelength grid ##
                # hdu_e2ds=fits.open('%s'%file.replace('s1d', 'e2ds'))
                # spec_e2ds=hdu_e2ds[0].data
                # header_e2ds=hdu_e2ds[0].header

                # wave_e2ds=get_wave(spec_e2ds, header_e2ds)*(1.+brv/2.99792458e5)

                # # plt.figure()
                # # plt.scatter(np.arange(len(wave_e2ds[order][:-1])), wave_e2ds[order][1:]-wave_e2ds[order][:-1], label = 'e2ds wave (after berv)')
                # # plt.scatter(np.arange(len(wave_e2ds[order][:-1])), get_wave(spec_e2ds, header_e2ds)[order][1:]-get_wave(spec_e2ds, header_e2ds)[order][:-1], label = 'e2ds wave (before berv)')
                # # # plt.scatter(np.arange(len(wavelengths[:-1])), wavelengths[:-1]-wavelengths[1:], label = 's1d wave')
                # # plt.legend()
                # # plt.show()

                # # id = np.logical_and(wave_e2ds<np.max(wavelengths), wave_e2ds>np.min(wavelengths))
                # # print(wave_e2ds*u.AA)
                # # print(wavelengths*u.AA)
                # # print(fluxes*u.photon)

                # blaze_file = glob.glob('%sblaze_folder/**blaze_A*.fits'%(directory))
                # # print('%sblaze_folder/**blaze_A*.fits'%(directory))
                # # print(blaze_file)
                # blaze_file = blaze_file[0]
                # blaze =fits.open('%s'%blaze_file)
                # blaze_func = blaze[0].data
                # spec_e2ds = spec_e2ds/blaze_func
            
                # diff_arr = wavelengths[1:] - wavelengths[:-1]
                # print(diff_arr)
                # wavelengths = wavelengths[:-1]
                # fluxes = fluxes[:-1]/diff_arr

                # s1d_spec = Spectrum1D(spectral_axis = wavelengths*u.AA, flux = fluxes*u.Unit('photon AA-1'))
                # fluxcon = FluxConservingResampler()
                # new_spec = fluxcon(s1d_spec, wave_e2ds[order]*u.AA)

                # wavelengths = new_spec.spectral_axis
                # fluxes = new_spec.flux

                # wavelengths = wavelengths[10:len(wave_e2ds[order])-9]/u.AA
                # fluxes = fluxes[10:len(wave_e2ds[order])-9]/u.Unit('photon AA-1')
                # flux_error_order = flux_error_order[10:len(wave_e2ds[order])-10]

                # diff_arr = wavelengths[1:] - wavelengths[:-1]
                # print(diff_arr)
                # wavelengths = wavelengths[:-1]
                # fluxes = fluxes[:-1]*diff_arr

                # print(wavelengths)
                # print(fluxes)

                # plt.figure()
                # plt.title('interpolated s1d comapred to actual e2ds spectrum')
                # plt.plot(wavelengths, fluxes, label = 'interpolated s1d on e2ds wave grid')
                # plt.plot(wave_e2ds[order], spec_e2ds[order], label = 'e2ds spectrum')
                # plt.legend()
                # plt.show()


                # ## end of test ##

                # ## test2 - synthetic s1d spectrum - using wavelength grid ##
                # def gauss(x1, rv, sd, height, cont):
                #     y1 = height*np.exp(-(x1-rv)**2/(2*sd**2)) + cont
                #     return y1

                # wavelength_grid = wavelengths
                # flux_grid = np.ones(wavelength_grid.shape)

                # linelist = '/home/lsd/Documents/fulllinelist0001.txt'
                
                # linelist_expected = np.genfromtxt('%s'%linelist, skip_header=4, delimiter=',', usecols=(1,9))
                # wavelengths_expected1 =np.array(linelist_expected[:,0])
                # depths_expected1 = np.array(linelist_expected[:,1])
                # # print(len(depths_expected1))

                # wavelength_min = np.min(wavelengths)
                # wavelength_max = np.max(wavelengths)

                # print(wavelength_min, wavelength_max)

                # wavelengths_expected=[]
                # depths_expected=[]
                # no_line =[]
                # for some in range(0, len(wavelengths_expected1)):
                #     line_min = 0.25
                #     if wavelengths_expected1[some]>=wavelength_min and wavelengths_expected1[some]<=wavelength_max and depths_expected1[some]>=line_min:
                #         wavelengths_expected.append(wavelengths_expected1[some])
                #         #depths_expected.append(depths_expected1[some]+random.uniform(-0.1, 0.1))
                #         depths_expected.append(depths_expected1[some])
                #     else:
                #         pass
                
                # count_range = np.array([len(wavelengths_expected)]*10)*np.array([0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9])
                # count_range = np.array(count_range, dtype = int)
                # print(count_range)
                # vgrid = np.linspace(-21,18,48)
                # try: ccf = fits.open(file.replace('s1d', 'ccf_K5'))
                # except: ccf = fits.open(file.replace('s1d', 'ccf_G2'))
                # rv = ccf[0].header['ESO DRS CCF RV']

                
                # for line in count_range:
                #     mid_wave = wavelengths_expected[line]
                #     wgrid = 2.99792458e5*mid_wave/(2.99792458e5-vgrid)
                #     id = np.logical_and(wavelength_grid<np.max(wgrid), wavelength_grid>np.min(wgrid))
                #     prof_wavelength_grid = wavelength_grid[id]
                #     prof_v_grid = ((prof_wavelength_grid - mid_wave)*2.99792458e5)/prof_wavelength_grid
                #     prof = gauss(prof_v_grid, rv, 2.47, -depths_expected[line], 1.)
                #     # plt.figure()
                #     # plt.plot(prof_wavelength_grid, prof)
                #     # plt.show()
                #     flux_grid[id] = prof

                # coeffs=np.polyfit(wavelengths, fluxes/fluxes[0], 3)
                # poly = np.poly1d(coeffs*fluxes[0])
                # fit = poly(wavelengths)

                # wavelengths = wavelength_grid
                # fluxes = flux_grid * fit

# Old blaze correct code part 2:

# plt.figure()
            # plt.figure('blaze for orders 28, 29 and 30')
            # plt.plot(wave[28], blaze[0].data[28])
            # plt.plot(wave[29], blaze[0].data[29])
            # plt.plot(wave[30], blaze[0].data[30])

            # plt.figure()
            # plt.title('blaze for orders 28, 29 and 30 summed together')
            # pixel_grid = np.linspace(np.min(wave[28]), np.max(wave[30]), len(np.unique(wave[28:30])))
            # blaze_sum = np.zeros(pixel_grid.shape)

            # f28 = interp1d(wave[28], blaze[0].data[28], kind = 'linear', bounds_error=False, fill_value = 'extrapolate')
            # f29 = interp1d(wave[29], blaze[0].data[29], kind = 'linear', bounds_error=False, fill_value = 'extrapolate')
            # f30 = interp1d(wave[30], blaze[0].data[30], kind = 'linear', bounds_error=False, fill_value = 'extrapolate')

            # idx28 = np.logical_and(pixel_grid>np.min(wave[28]), pixel_grid<np.max(wave[28]))
            # idx29 = np.logical_and(pixel_grid>np.min(wave[29]), pixel_grid<np.max(wave[29]))
            # idx30 = np.logical_and(pixel_grid>np.min(wave[30]), pixel_grid<np.max(wave[30]))

            # blaze_28 = f28(pixel_grid[idx28])
            # blaze_29 = f29(pixel_grid[idx29])
            # blaze_30 = f30(pixel_grid[idx30])

            # wave28 = pixel_grid[idx28]
            # wave29 = pixel_grid[idx29]
            # wave30 = pixel_grid[idx30]

            # for pixel in range(len(pixel_grid)):
            #     wavell = pixel_grid[pixel]
                
            #     idx28 = tuple([wave28==wavell])
            #     idx29 = tuple([wave29==wavell])
            #     idx30 = tuple([wave30==wavell])
                
            #     try: b = blaze_28[idx28][0]
            #     except: b=0
            #     try: b1 = blaze_29[idx29][0]
            #     except: b1=0
            #     try: b2 = blaze_30[idx30][0]
            #     except: b2=0

            #     print(wavell, b, b1, b2)

            #     blaze_sum[pixel] = b + b1 + b2 

            # plt.plot(pixel_grid, blaze_sum)
            # plt.show()

            # plt.figure()
            # plt.title('e2ds after blaze orders 28, 29 and 30')
            # plt.plot(wave[28], spec[28])
            # plt.plot(wave[29], spec[29])
            # plt.plot(wave[30], spec[30])
            
            ## TEST - adjusting e2ds spectrum onto s1d continuum ##

            # ## first interpolate s1d onto e2ds wavelength grid ##
            # s1d_file = fits.open(file.replace('e2ds', 's1d'))
            # s1d_spec = s1d_file[0].data
            # wave_s1d = s1d_file[0].header['CRVAL1']+(np.arange(s1d_spec.shape[0]))*s1d_file[0].header['CDELT1']

            # wavelengths = wave_s1d
            # fluxes = s1d_spec

            # plt.figure()
            # plt.plot(wavelengths, fluxes, label = 'e2ds spectrum - corrected to s1d continuum')
            # plt.plot(wave_s1d, s1d_spec, label = 's1d spectrum')
            # plt.legend()
            # plt.show()

            # plt.figure()
            # plt.plot(wavelengths, fluxes, label = 's1d')
            # plt.legend()

            # diff_arr = wavelengths[1:] - wavelengths[:-1]
            # # print(diff_arr)
            # wavelengths = wavelengths[:-1]
            # fluxes = fluxes[:-1]

            # fluxes = fluxes/diff_arr

            # fluxes = np.ones(fluxes.shape)
            # for i in range(len(fluxes)):
            #     fluxes[i] = fluxes[i]*8000
            # plt.figure()
            # plt.plot(wavelengths, fluxes, label = 's1d in photons AA-1')
            # #plt.legend()

            # interpolate s1d onto e2ds wavlengths - non flux conserving
            # s1dd_spec = Spectrum1D(spectral_axis = wavelengths*u.AA, flux = fluxes*u.Unit('photon AA-1'))
            # fluxcon = FluxConservingResampler()
            # extended_e2ds_wave = np.concatenate((wave[order], [wave[order][-1]+0.01]))

            ## MM-LSD way

            # fluxes = spec[order]
            # flux_error_order = flux_error[order]
            # wavelengths = wave[order]

            # for i in range(len(fluxes)):
            #     if i ==0:
            #         # print(fluxes[i])
            #         fluxes[i] = fluxes[i]*(0.01/(wavelengths[1]-wavelengths[0]))
            #         # print(0.01/(2.99792458e5*(wavelengths[1]-wavelengths[0])))
            #         # print(fluxes[i])
            #     else:
            #         # print(fluxes[i])
            #         fluxes[i] = fluxes[i]*(0.01/(wavelengths[i]-wavelengths[i-1]))
            #         # print(0.01/(2.99792458e5*(wavelengths[1]-wavelengths[0])))
            #         # print(fluxes[i])

            # ## end of MM-LSD way

            # new_spec = fluxcon(s1dd_spec, wavelengths*u.AA)

            # reference_wave = new_spec.spectral_axis/u.AA
            # reference_flux = new_spec.flux/u.Unit('photon AA-1')

            # # print(len(wavelengths_new))
            # # print(len(wave[order]))

            # # # plt.plot(wavelengths_new, fluxes, label = 'interpolated s1d in photons AA-1')
            # # # plt.legend()

            # #diff_arr = wavelengths_new[1:] - wavelengths_new[:-1]
            # #reference_wave = wavelengths_new[:-1]
            # #reference_flux = fluxes[:-1]*diff_arr
            
            # # # print(len(reference_wave))
            # # # print(len(wave[order]))

            # # # print(reference_wave-wave[order])

            # # # plt.figure()
            # # # plt.plot(reference_wave, reference_flux, label = 'interpolated s1d in photons per bin')
            # # # plt.legend()
            # # # plt.show()
            # # ## divide e2ds spectrum by interpolated s1d and fit polynomial to result

            # reference_wave = np.array(reference_wave, dtype = float)
            # reference_flux = np.array(reference_flux, dtype = float)
            # div_frame = fluxes/reference_flux

            # # plt.figure()
            # # plt.plot(reference_wave, div_frame)
            # # plt.show()

            # # # ### creating windows to fit polynomial to
            # # # binned = np.zeros(int(len(div_frame)/2))
            # # # binned_waves = np.zeros(int(len(div_frame)/2))
            # # # for i in range(0, len(div_frame)-1, 2):
            # # #     pos = int(i/2)
            # # #     binned[pos] = (div_frame[i]+div_frame[i+1])/2
            # # #     binned_waves[pos] = (reference_wave[i]+reference_wave[i+1])/2

            # # # plt.plot(frame_wavelengths[n], frames_unadjusted[n], color = 'b', label = 'unadjusted')
            # # # plt.figure()
            # # # plt.plot(frame_wavelengths[n], frames[n])
            # # # plt.show()

            # ### fitting polynomial to div_frame
            # coeffs=np.polyfit(reference_wave, div_frame, 3)
            # poly = np.poly1d(coeffs)
            # # print(coeffs)
            # inputs = coeffs[::-1]
            # # print(inputs)

            # wavelengths = reference_wave

            # fit = poly(wavelengths)
        
            # # # plt.figure()
            # # # plt.plot(reference_wave, reference_flux, label= 'reference')
            # # # plt.plot(wave[order], spec[order], label = 'e2ds')
            # # # plt.legend()

            # # # plt.figure()
            # # # plt.plot(reference_wave-wavelengths[:-1], label = 'reference_wave-wavelengths')
            # # # plt.legend()
            # # # plt.show()

            # # plt.figure()
            # # plt.scatter(reference_wave, div_frame, label = 'div flux')
            # # plt.plot(wavelengths, fit, label = 'fit')
            # # # plt.plot(reference_wave, poly(reference_wave), label = 'poly(reference_wave)')
            # # plt.legend()
            # # plt.show()
            
            # fluxes = spec[order]/fit
            # flux_error_order = flux_error[order]/fit

            # # plt.figure()
            # # plt.plot(wavelengths, fluxes/reference_flux, label = 'continuum adjusted e2ds/s1d')
            # # plt.legend()
            # # plt.show()

            # # plt.figure()
            # # plt.plot(wavelengths, fluxes-reference_flux, label = 'continuum adjusted e2ds-s1d')
            # # plt.legend()
            # # plt.show()
            # # plt.figure()
            # # plt.plot(wavelengths, spec[order], label = 'before')
            # # plt.plot(wavelengths, fit)
            # # plt.plot(wavelengths, fluxes, label= 'after')
            # # plt.legend()
            # # plt.show()

            # # idx_full = np.logical_and(wave_s1d>np.min(wave[28]), wave_s1d<np.max(wave[30]))
            # # plt.plot(wave_s1d[idx_full], s1d_spec[idx_full])
            # # plt.show()

            # blaze.close()
            # # plt.figure('after blaze correction - e2ds vs s1d - after berv correction')
            # # plt.plot(wave[order], spec[order], label = 'e2ds')
            # # plt.plot(wave_s1d[wave_s1d>np.max(wave[order])], spec_s1d[wave_s1d>np.max(wave[order])], label = 's1d')
            # # plt.show()

            

            # # # test - e2ds interpolated onto s1d wavelength grid ##
            # # hdu_s1d=fits.open('%s'%file.replace('e2ds', 's1d'))
            # # spec_s1d=hdu_s1d[0].data
            # # header_s1d=hdu_s1d[0].header

            # # wave_s1d=header_s1d['CRVAL1']+(header_s1d['CRPIX1']+np.arange(spec_s1d.shape[0]))*header_s1d['CDELT1']
            # # id = np.logical_and(wave_s1d<np.max(wavelengths), wave_s1d>np.min(wavelengths))
            # # print(wave_s1d*u.AA)
            # # print(wavelengths*u.AA)
            # # print(fluxes*u.Unit('erg cm-2 s-1 AA-1'))
            # # # plt.figure('s1d compared to interpolated e2ds')
            # # # plt.title('s1d compared to interpolated e2ds')
            # # # plt.plot(wave_s1d, spec_s1d, label = 's1d spectrum')

            # # ## these fluxes are in photons per bin - I need them in photons per Angstrom
            # # ## therefore i do flux/angstroms in pixel
            # # diff_arr = wavelengths[1:] - wavelengths[:-1]
            # # print(diff_arr)
            # # wavelengths = wavelengths[:-1]
            # # fluxes = fluxes[:-1]
            # # # plt.figure('changing flux units')
            # # # plt.plot(wavelengths, fluxes, label = 'flux per pixel')

            # # fluxes = fluxes/diff_arr

            # # # plt.plot(wavelengths, fluxes, label = ' flux per A')

            # # e2ds_spec = Spectrum1D(spectral_axis = wavelengths*u.AA, flux = fluxes*u.Unit('photon AA-1'))
            # # fluxcon = FluxConservingResampler()
            # # new_spec = fluxcon(e2ds_spec, wave_s1d[id]*u.AA)

            # # wavelengths = new_spec.spectral_axis
            # # fluxes = new_spec.flux

            # # wavelengths = wavelengths[:4097]/u.AA
            # # fluxes = fluxes[:4097]/u.Unit('photon AA-1')

            # # diff_arr = wavelengths[1:] - wavelengths[:-1]
            # # wavelengths = wavelengths[:-1]
            # # fluxes = fluxes[:-1]*diff_arr

            # # # plt.figure('s1d compared to interpolated e2ds') 
            # # # plt.plot(wavelengths, fluxes, label = 'interpolated e2ds onto s1d')
            # # # plt.xlim(np.min(wavelengths), np.max(wavelengths))
            # # # plt.legend()
            # # # plt.show() 

            # # print(wavelengths)
            # # print(fluxes)

            # #end of test ##

            # # # test2 - synthetic e2ds spectrum - using wavelength grid ##
            # # def gauss(x1, rv, sd, height, cont):
            # #     y1 = height*np.exp(-(x1-rv)**2/(2*sd**2)) + cont
            # #     return y1

            # # wavelength_grid = wavelengths
            # # flux_grid = np.ones(wavelength_grid.shape)

            # # linelist = '/home/lsd/Documents/fulllinelist0001.txt'

            # # linelist_expected = np.genfromtxt('%s'%linelist, skip_header=4, delimiter=',', usecols=(1,9))
            # # wavelengths_expected1 =np.array(linelist_expected[:,0])
            # # depths_expected1 = np.array(linelist_expected[:,1])
            # # # print(len(depths_expected1))

            # # wavelength_min = np.min(wavelengths)
            # # wavelength_max = np.max(wavelengths)

            # # print(wavelength_min, wavelength_max)

            # # wavelengths_expected=[]
            # # depths_expected=[]
            # # no_line =[]
            # # for some in range(0, len(wavelengths_expected1)):
            # #     line_min = 0.25
            # #     if wavelengths_expected1[some]>=wavelength_min and wavelengths_expected1[some]<=wavelength_max and depths_expected1[some]>=line_min:
            # #         wavelengths_expected.append(wavelengths_expected1[some])
            # #         #depths_expected.append(depths_expected1[some]+random.uniform(-0.1, 0.1))
            # #         depths_expected.append(depths_expected1[some])
            # #     else:
            # #         pass

            # # count_range = np.array([len(wavelengths_expected)]*10)*np.array([0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9])
            # # count_range = np.array(count_range, dtype = int)
            # # print(count_range)
            # # vgrid = np.linspace(-21,18,48)
            # # try: ccf = fits.open(file.replace('e2ds', 'ccf_K5'))
            # # except: ccf = fits.open(file.replace('e2ds', 'ccf_G2'))
            # # rv = ccf[0].header['ESO DRS CCF RV']


            # # for line in count_range:
            # #     mid_wave = wavelengths_expected[line]
            # #     wgrid = 2.99792458e5*mid_wave/(2.99792458e5-vgrid)
            # #     id = np.logical_and(wavelength_grid<np.max(wgrid), wavelength_grid>np.min(wgrid))
            # #     prof_wavelength_grid = wavelength_grid[id]
            # #     prof_v_grid = ((prof_wavelength_grid - mid_wave)*2.99792458e5)/prof_wavelength_grid
            # #     prof = gauss(prof_v_grid, rv, 2.47, -depths_expected[line], 1.)
            # #     # id = tuple([prof_v_grid<-0.99])
            # #     # plt.figure()
            # #     # plt.plot(prof_wavelength_grid, prof)
            # #     # plt.show()
            # #     flux_grid[id] = prof

            # # # plt.figure()
            # # # plt.plot(wavelength_grid, flux_grid)
            # # # plt.show()

            # # coeffs=np.polyfit(wavelengths, fluxes/fluxes[0], 3)
            # # poly = np.poly1d(coeffs*fluxes[0])
            # # fit = poly(wavelengths)

            # # wavelengths = wavelength_grid
            # # fluxes = flux_grid * fit

            # # plt.figure()
            # plt.plot(wavelengths, fluxes)
            # plt.show()

            # find overlapping regions 
            # last_wavelengths = wave[order-1]
            # next_wavelengths = wave[order+1]
            # last_spec = spec[order-1]
            # next_spec = spec[order+1]
            # last_error = flux_error[order-1]
            # next_error = flux_error[order+1]
            # min_overlap = np.min(wavelengths)
            # max_overlap = np.max(wavelengths)

            
            # idx_ = tuple([wavelengths>min_overlap])
            # last_idx = np.logical_and(last_wavelengths>min_overlap, last_wavelengths<max_overlap)
            # next_idx = np.logical_and(next_wavelengths>min_overlap, next_wavelengths<max_overlap)
            
            # overlap = np.array(([list(last_wavelengths[last_idx]), list(last_spec[last_idx]), list(last_error[last_idx])], [list(next_wavelengths[next_idx]), list(next_spec[next_idx]), list(next_error[next_idx])]))
            
            # # overlap[0, 0] = list(last_wavelengths[last_idx])
            # # overlap[0, 1] = list(last_spec[last_idx])
            # # overlap[1, 0] = list(next_wavelengths[next_idx])
            # # overlap[1, 1] = list(next_spec[next_idx])

            # print(overlap)
            # plt.figure()
            # plt.plot(wavelengths, fluxes)
            # plt.plot(wavelengths[idx_overlap], fluxes[idx_overlap])
            # plt.show()
            # wavelengths = wavelengths[idx]
            # fluxes = fluxes[idx]
            # flux_error_order = flux_error_order[idx]

# Telluric correction:
## telluric correction
        # tapas = fits.open('/Users/lucydolan/Starbase/tapas_000001.fits')
        # tapas_wvl = (tapas[1].data["wavelength"]) * 10.0
        # tapas_trans = tapas[1].data["transmittance"]
        # tapas.close()
        # brv=header['ESO DRS BERV']
        # tapas_wvl = tapas_wvl[::-1]/(1.+brv/2.99792458e5)
        # tapas_trans = tapas_trans[::-1]

        # background = upper_envelope(tapas_wvl, tapas_trans)
        # f = interp1d(tapas_wvl, tapas_trans / background, bounds_error=False)

        # plt.figure('telluric spec and real spec')
        # plt.plot(wavelengths, continuumfit(wavelengths, fluxes, 3))
        # plt.plot(wavelengths, f(wavelengths))
        # plt.show()
        
        # plt.figure()
        # plt.plot(tapas_wvl, tapas_trans)
        # plt.show()
        # print('overlap accounted for')

# Overlap stuff?
# print(len(wavelengths))
                # print(np.max(wavelengths), np.min(wavelengths))
                # print(min_overlap)
                # print(max_overlap)
                # idx_overlap = np.logical_and(wavelengths>=min_overlap, wavelengths<=max_overlap)
                # idx_overlap = tuple([idx_overlap==False])
                # overlap = []

                # plt.figure()
                # plt.plot(wavelengths, fluxes, label = 'Flux')
                # # plt.plot(wavelengths[idx_overlap], fluxes[idx_overlap])
                # plt.show()

                # plt.plot(wavelengths, fluxes, label = 's1d')
                # plt.legend()
                # plt.show()


# OLD LSD FUNCTIONS

    def get_wave(self, data, header):

        wave = np.array(data*0., dtype = 'longdouble')
        no = data.shape[0]
        npix = data.shape[1]
        d = header['ESO DRS CAL TH DEG LL']
        xx0 = np.arange(npix)
        xx = []
        for i in range(d+1):
            xx.append(xx0**i)
        xx = np.asarray(xx, dtype = 'longdouble')

        for o in range(no):
            for i in range(d+1):
                idx = i + o * (d + 1)
                par = np.longdouble(header['ESO DRS CAL TH COEFF LL%d' % idx])
                wave[o, :] = wave[o, :] + par * xx[i, :]
            #for x in range(npix):
            #  wave[o,x]=wave[o,x]+par*xx[i,x]#float(x)**float(i)

        return wave

    def upper_envelope(self, x, y):
        # from MM-LSD code - give credit if needed
        # used to compute the tapas continuum. find peaks then fit spline to it.
        peaks = find_peaks(y, height=0.2, distance=len(x) // 500)[0]
        # t= knot positions
        spl = LSQUnivariateSpline(x=x[peaks], y=y[peaks], t=x[peaks][5::10])
        return spl(x)

    def blaze_correct(self, file_type, spec_type, order, file, directory, masking, run_name, berv_opt):
        #### Inputing spectrum depending on file_type and spec_type #####

        if file_type == 's1d':
            #### Finding min and max wavelength from e2ds for the specified order ######
            file_e2ds = file.replace('s1d', 'e2ds')

            hdu=fits.open('%s'%file_e2ds)
            sn = hdu[0].header['HIERARCH ESO DRS SPE EXT SN%s'%order]
            spec = hdu[0].data
            header = hdu[0].header
            brv = header['ESO DRS BERV']

            spec_check = spec[spec<=0]
            # if len(spec_check)>0:
                # print('WARNING NEGATIVE/ZERO FLUX - corrected')

            where_are_zeros = (spec<=0)
            spec[where_are_zeros] = 1e12
            flux_error = np.sqrt(spec)

            wave=self.get_wave(spec, header)*(1.+brv/2.99792458e5)
            wavelengths_order = wave[order]
            wavelength_min = np.min(wavelengths_order)
            wavelength_max = np.max(wavelengths_order)

            ## remove overlapping region (always remove the overlap at the start of the order, i.e the min_overlap)
            last_wavelengths = wave[order-1]
            next_wavelengths = wave[order+1]
            min_overlap = np.max(last_wavelengths)
            max_overlap = np.min(next_wavelengths)

            # idx_ = tuple([wavelengths>min_overlap])
            # wavelength_min = 5900
            # wavelength_max = wavelength_min+200    ###### if you want to do a WAVELENGTH RANGE just input min and max here ######
            hdu.close()

            #### Now reading in s1d file ########
            hdu=fits.open('%s'%file)
            spec=hdu[0].data
            header=hdu[0].header
            spec_check = spec[spec<=0]
            
            wave=hdu[0].header['CRVAL1']+(np.arange(spec.shape[0]))*hdu[0].header['CDELT1']
            
            where_are_zeros = (spec<=0)
            spec[where_are_zeros] = 1e12
            flux_error = np.sqrt(spec)

            if spec_type == 'order':
                wavelengths=[]
                fluxes=[]
                errors = []
                for value in range(0, len(wave)):
                    l_wave=wave[value]
                    l_flux=spec[value]
                    l_error=flux_error[value]
                    if l_wave>=wavelength_min and l_wave<=wavelength_max:
                        wavelengths.append(l_wave)
                        fluxes.append(l_flux)
                        errors.append(l_error)
                wavelengths = np.array(wavelengths)
                fluxes = np.array(fluxes)
                errors = np.array(errors)


                if len(wavelengths)>5144:
                    wavelengths = wavelengths[:5144]
                    fluxes = fluxes[:5144]
                    flux_error = errors[:5144]

                spec_check = fluxes[fluxes<=0]
                # if len(spec_check)>0:
                #     print('WARNING NEGATIVE/ZERO FLUX - corrected')

                # where_are_zeros = (spec<=0)
                # spec[where_are_zeros] = 1e12
                # flux_error = np.sqrt(spec)
                
                flux_error_order = flux_error
                masked_waves = []
                masked_waves = np.array(masked_waves)

                if masking == 'masked':
                    ## I've just been updating as I go through so it's not complete
                    #if you want to add to it the just add an element of form: [min wavelength of masked region, max wavelength of masked region]
                    masks_csv = np.genfromtxt('/home/lsd/Documents/HD189733b_masks.csv', delimiter=',')
                    min_waves_mask = np.array(masks_csv[:,0])
                    max_waves_mask = np.array(masks_csv[:,1])
                    masks = []
                    for mask_no in range(len(min_waves_mask)):
                        masks.append([min_waves_mask[mask_no], max_waves_mask[mask_no]])

                    masked_waves=[]

                    for mask in masks:
                        #print(np.max(mask), np.min(mask))
                        idx = np.logical_and(wavelengths>=np.min(mask), wavelengths<=np.max(mask))
                        #print(flux_error_order[idx])
                        flux_error_order[idx] = 1e12
                        #print(idx)
                        if len(wavelengths[idx])>0:
                            masked_waves.append(wavelengths[idx])

                    #masks = []
                    ### allows extra masking to be added ##

                    plt.figure('masking')
                    plt.plot(wavelengths, fluxes)
                    if len(masked_waves)>0:
                        for masked_wave in masked_waves:
                            plt.axvspan(np.min(masked_wave), np.max(masked_wave), alpha=0.5, color='red')
                    #plt.show()

                    #response = input('Are there any regions to be masked? y or n: ')
                    response = 'y'
                    if response == 'y':
                        '''
                        print('Take note of regions.')
                        plt.figure('masking')
                        plt.plot(wavelengths, fluxes)
                        plt.show()
                        '''
                        #response1 = int(input('How many regions to mask?: '))
                        response1 = 0
                        for i in range(response1):
                            min_wave = float(input('Minimum wavelength of region %s: '%i))
                            max_wave = float(input('Maximum wavelength of region %s: '%i))
                            masks.append([min_wave, max_wave])
                        masked_waves=[]
                        #print(masks)
                        for mask in masks:

                            idx = np.logical_and(wavelengths>=np.min(mask), wavelengths<=np.max(mask))

                            flux_error_order[idx] = 1e12

                            if len(wavelengths[idx])>0:
                                masked_waves.append(wavelengths[idx])

                        plt.figure('telluric masking')
                        plt.title('Spectrum - after telluric masking')
                        plt.plot(wavelengths, fluxes)
                        for masked_wave in masked_waves:
                            plt.axvspan(np.min(masked_wave), np.max(masked_wave), alpha=0.5, color='red')
                        plt.savefig('/home/lsd/Documents/LSD_Figures/masking_plots/order%s_masks_%s'%(order, run_name))

                        plt.figure('errors')
                        plt.plot(wavelengths, flux_error_order)
                        plt.close('all')
                        #plt.show()

                elif masking == 'unmasked':
                    masked_waves = []
                    masked_waves = np.array(masked_waves)

            elif spec_type == 'full':
                ## not set up properly.
                wavelengths = wave
                fluxes = spec

        elif file_type == 'e2ds':
            hdu=fits.open('%s'%file)
            spec=hdu[0].data
            header=hdu[0].header
            sn = hdu[0].header['HIERARCH ESO DRS SPE EXT SN%s'%order]

            spec_check = spec[spec<=0]
            # if len(spec_check)>0:
            #     print('WARNING NEGATIVE/ZERO FLUX - corrected')


            # where_are_NaNs = np.isnan(flux_error)
            # flux_error[where_are_NaNs] = 1e12
            where_are_zeros = (spec<=0)
            spec[where_are_zeros] = 1e12
            flux_error = np.sqrt(spec)
            '''
            flux_error1 = header['HIERARCH ESO DRS SPE EXT SN%s'%order]

            flux_error = header['HIERARCH ESO DRS CAL TH RMS ORDER%s'%order]
            print(flux_error, flux_error1)

            flux_error = flux_error*np.ones(np.shape(spec))
            '''
            # file_ccf = fits.open(file.replace('e2ds', 'ccf_G2'))
            # print(file_ccf[0].header['ESO DRS BERV'])
            brv = np.longdouble(header['ESO DRS BERV'])
            # print(brv)
            wave_nonad = self.get_wave(spec, header)
            # if berv_opt == 'y':
            #     print('BERV corrected')
            wave = wave_nonad*(1.+brv/2.99792458e5)
            wave = np.array(wave, dtype = 'double')
            # if berv_opt == 'n':
            #     print('BERV not corrected')
            # wave = wave_nonad
        
            # rv_drift=header['ESO DRS DRIFT RV'] 
            # print(rv_drift)
            wave_corr = (1.+brv/2.99792458e5)
            # print(brv, (wave_corr-1)*2.99792458e5)

            # inp = input('Enter to continue...')
            '''
            plt.figure('Spectrum directly from fits file')
            plt.title('Spectrum directly from fits file')
            plt.errorbar(wave[order], spec[order], yerr = flux_error[order])
            plt.xlabel('wavelengths')
            plt.ylabel('flux')
            plt.show()
            '''

            blaze_file = glob.glob('tests/data/*blaze_B*.fits')

            blaze_file = blaze_file[0]
            # try:
            #     blaze_file = glob.glob('./**blaze_A*.fits')
            #     # print('%sblaze_folder/**blaze_A*.fits'%(directory))
            #     # print(blaze_file)
            #     blaze_file = blaze_file[0]
            # except: 
            #     try:
            #         blaze_file = glob.glob('/Users/lucydolan/Starbase/problem_frames/**blaze_A*.fits')
            #         blaze_file = blaze_file[0]
            #     except:
            #         blaze_file = glob.glob('tests/data/**blaze_A*.fits')
            blaze = fits.open('%s'%blaze_file)
            blaze_func = blaze[0].data
            min_rows = min(spec.shape[0], blaze_func.shape[0], flux_error.shape[0])
            spec, blaze_func, flux_error = spec[:min_rows], blaze_func[:min_rows], flux_error[:min_rows]
            spec = spec/blaze_func
            flux_error = flux_error/blaze_func

            fluxes = spec[order]
            flux_error_order = flux_error[order]
            wavelengths = wave[order]

            if masking == 'masked':
                ## I've just been updating as I go through so it's not complete
                #if you want to add to it the just add an element of form: [min wavelength of masked region, max wavelength of masked region]
                masks_csv = np.genfromtxt('/home/lsd/Documents/HD189733b_masks.csv', delimiter=',')
                min_waves_mask = np.array(masks_csv[:,0])
                max_waves_mask = np.array(masks_csv[:,1])
                masks = []
                for mask_no in range(len(min_waves_mask)):
                    masks.append([min_waves_mask[mask_no], max_waves_mask[mask_no]])

                masked_waves=[]

                for mask in masks:
                    #print(np.max(mask), np.min(mask))
                    idx = np.logical_and(wavelengths>=np.min(mask), wavelengths<=np.max(mask))
                    #print(flux_error_order[idx])
                    flux_error_order[idx] = 1e12
                    #print(idx)
                    if len(wavelengths[idx])>0:
                        masked_waves.append(wavelengths[idx])

                #masks = []
                ### allows extra masking to be added ##

                plt.figure('masking')
                plt.plot(wavelengths, fluxes)
                if len(masked_waves)>0:
                    for masked_wave in masked_waves:
                        plt.axvspan(np.min(masked_wave), np.max(masked_wave), alpha=0.5, color='red')
                #plt.show()

                #response = input('Are there any regions to be masked? y or n: ')
                response = 'y'
                if response == 'y':
                    '''
                    print('Take note of regions.')
                    plt.figure('masking')
                    plt.plot(wavelengths, fluxes)
                    plt.show()
                    '''
                    #response1 = int(input('How many regions to mask?: '))
                    response1 = 0
                    for i in range(response1):
                        min_wave = float(input('Minimum wavelength of region %s: '%i))
                        max_wave = float(input('Maximum wavelength of region %s: '%i))
                        masks.append([min_wave, max_wave])
                    masked_waves=[]
                    #print(masks)
                    for mask in masks:
                        #print(np.max(mask), np.min(mask))
                        idx = np.logical_and(wavelengths>=np.min(mask), wavelengths<=np.max(mask))
                        #print(flux_error_order[idx])
                        flux_error_order[idx] = 1e12

                        if len(wavelengths[idx])>0:
                            masked_waves.append(wavelengths[idx])

                    plt.figure('Telluric masking')
                    plt.title('Spectrum - telluric masking')
                    plt.plot(wavelengths, fluxes)
                    # print(masked_waves)
                    for masked_wave in masked_waves:
                        plt.axvspan(np.min(masked_wave), np.max(masked_wave), alpha=0.5, color='red')
                    #print('new version')
                    plt.savefig('/home/lsd/Documents/LSD_Figures/masking_plots/order%s_masks_%s'%(order, run_name))
                    plt.ylabel('flux')
                    plt.xlabel('wavelength')

                if response == 'n':
                    print('yay!')


            elif masking == 'unmasked':
                masked_waves = []
                masked_waves = np.array(masked_waves)

            else: print('WARNING - masking not catching - must be either "masked" or "unmasked"')

            hdu.close()

        #flux_error_order = (flux_error_order)/(np.max(fluxes)-np.min(fluxes))
        #print('flux error: %s'%flux_error_order)
        #fluxes = (fluxes - np.min(fluxes))/(np.max(fluxes)-np.min(fluxes))
        #idx = tuple([fluxes>0])

        return np.array(fluxes), np.array(wavelengths), np.array(flux_error_order), sn ## for just LSD



"""
All of the utility functions for the ACID package. Some functions may not be useful to the user.
"""
from __future__ import annotations
from beartype import beartype
from beartype.vale import IsAttr, IsEqual
import numpy as np
import glob, emcee, psutil, os
from emcee import EnsembleSampler
import emcee.backends.backend as emceebackend
import scipy.constants as const
from typing import TypeAlias, Annotated
c_kms = float(const.c/1e3)
FloatLike: TypeAlias = float | np.floating
IntLike: TypeAlias = int | np.integer
Scalar: TypeAlias = FloatLike | IntLike | Annotated[np.ndarray, IsAttr["size", IsEqual[1]]]
Array1D: TypeAlias = Annotated[np.ndarray, IsAttr["ndim", IsEqual[1]]] | list[Scalar]
Array2D: TypeAlias = Annotated[np.ndarray, IsAttr["ndim", IsEqual[2]]] | list[list[Scalar]] | list[Array1D]
Array3D: TypeAlias = Annotated[np.ndarray, IsAttr["ndim", IsEqual[3]]] | list[list[list[Scalar]]] | list[list[Array1D]] | list[Array2D]

def convert_moves_to_emcee(moves:list[tuple]):
    """Converts a list of move specifications to emcee moves.

    Parameters
    ----------
    moves : list[tuple], optional
        A list of tuples specifying the moves for the MCMC sampler. The format
        tries to follow the emcee documentation as closely as possible.
        However, the config cannot store classes directly, so move names are
        used instead and converted when building the sampler.

        Each tuple should have the form::

            (move_name: str, fraction: float, move_kwargs: dict | None)

        where:

        - "move_name" is the name of the emcee move. Supported variants currently
          include "RedBlueMove", "StretchMove", "WalkMove",
          "KDEMove", "DEMove", "DESnookerMove", "MHMove",
          and "GaussianMove". Refer to the emcee documentation for more
          details on each move type. Input move names are checked against the
          "emcee.moves" module, so other moves from that module may also work,
          although not all have been tested with ACID.
        - "fraction" is the fraction of walkers to which this move should be applied.
        - "move_kwargs" is an optional dictionary of keyword arguments passed to
          the move class initialisation.

    Returns
    -------
    list
        A list of emcee move objects corresponding to the input specifications.
    """
    emcee_moves = []
    for move in moves:
        if len(move) == 2:
            move_name, fraction = move
            move_kwargs = {}
        elif len(move) == 3:
            move_name, fraction, move_kwargs = move
            if not isinstance(move_kwargs, dict):
                raise ValueError(
                    "Move kwargs must be a dictionary of keyword arguments to pass to " \
                    "the move class initialisation (if passed)."
                )
        else:
            raise ValueError(
                "Each move tuple must have length 2 or 3: " \
                "(move_name, fraction) or (move_name, fraction, kwargs).")

        if not hasattr(emcee.moves, move_name):
            raise ValueError(f"Move '{move_name}' is not a valid emcee move.")
        move_class = getattr(emcee.moves, move_name)
        emcee_moves.append((move_class(**move_kwargs), fraction))
        
    return emcee_moves

def mask_invalid(wavelengths, flux, errors=None, return_mask=False, verbose=2):
    """Masks any pixels where the wavelength, flux, or error is infinite or <= 0.
    Replaces bad pixels with NaN, which ACID can handle.

    Parameters
    ----------
    wavelengths : array_like
        The wavelength values of the spectrum.
    flux : array_like
        The flux values of the spectrum.
    errors : array_like
        The error values associated with the spectrum.

    Returns
    -------
    tuple
        A tuple containing the cleaned wavelength, flux, and error arrays.
    """
    mask = (
        np.isfinite(wavelengths) &
        np.isfinite(flux) &
        (np.isfinite(errors) if errors is not None else True) &
        (flux > 0) &
        (errors > 0 if errors is not None else True)
    )
    fill_value = np.nan

    w = np.where(mask, wavelengths, fill_value)
    f = np.where(mask, flux, fill_value)
    e = np.where(mask, errors, fill_value) if errors is not None else None

    if verbose > 1:
        num_invalid = np.size(wavelengths) - np.count_nonzero(mask)
        perc_invalid = num_invalid / np.size(wavelengths) * 100
        if perc_invalid > 10:
            print(f"Your spectrum includes {num_invalid} out of {np.size(wavelengths)} non-positive/non-finite/nan values ({perc_invalid:.2f}%), \n"
                  f"which will be dropped when necessary, but it is still recommended to check your wavelength, \n"
                  f"spectrum and error arrays for bad pixels and make sure this is intentional. \n"
                  f"This warning is only printed if more than 10% of pixels are invalid.")

    output = (w, f, e) if errors is not None else (w, f)
    output = output + (mask,) if return_mask else output
    return output

def drop_invalid(wavelengths, flux, errors=None, return_mask=False, verbose=2):
    """Drops any pixels where the wavelength, flux, or error is infinite or <= 0.

    Parameters
    ----------
    wavelengths : array_like
        The wavelength values of the spectrum.
    flux : array_like
        The flux values of the spectrum.
    errors : array_like
        The error values associated with the spectrum.

    Returns
    -------
    tuple
        A tuple containing the cleaned wavelength, flux, and error arrays.
    """
    mask = (
        np.isfinite(wavelengths) &
        np.isfinite(flux) &
        (np.isfinite(errors) if errors is not None else True) &
        (flux > 0) &
        ((errors > 0) if errors is not None else True)
    )
    w = wavelengths[mask]
    f = flux[mask]
    e = errors[mask] if errors is not None else None

    if verbose > 1:
        num_invalid = np.size(wavelengths) - np.count_nonzero(mask)
        perc_invalid = num_invalid / np.size(wavelengths) * 100
        if perc_invalid > 10:
            print(f"Dropped {num_invalid} invalid pixels out of {np.size(wavelengths)} (non-finite or <= 0), which is {perc_invalid:.2f}% of the total.")

    output = (w, f, e) if errors is not None else (w, f)
    output = output + (mask,) if return_mask else output
    return output

def clip_wavelengths(wavelengths, wavelengths_linelist, depths_linelist, pad=5):
    """
    Clips the linelist to only include lines within the wavelength range of the observed spectrum.
    Includes a pad either side of the wavelength range so that the wings of lines outside
    the range can also contribute to the fit.

    Parameters
    ----------
    wavelengths : np.ndarray
        Wavelengths of the observed spectrum
    wavelengths_linelist : np.ndarray
        Wavelengths from the linelist
    depths_linelist : np.ndarray
        Depths from the linelist
    pad : float, optional
        Number of angstroms to pad on either side of the wavelength range. By default, 5.

    Returns
    -------
    wavelengths_linelist : np.ndarray
        Clipped wavelengths from the linelist
    depths_linelist : np.ndarray
        Clipped depths from the linelist
    """
    lower, upper = np.nanmin(wavelengths)-pad, np.nanmax(wavelengths)+pad
    idx = (wavelengths_linelist >= lower) & (wavelengths_linelist <= upper)
    return wavelengths_linelist[idx], depths_linelist[idx]

def drop_edges(array, n_pix=2):
    """
    Drops the edges of an array by a specified number of pixels.
    
    Parameters
    ----------
    array : np.ndarray
        The input array.
    n_pix : int, optional
        Number of pixels to drop from each edge. Default is 2.

    Returns
    -------
    np.ndarray
        The array with edges dropped.
    """
    return array[n_pix:-n_pix]

@beartype
def calc_deltav(wavelengths:Array1D)->Scalar:
    """Calculates velocity pixel size

    Calculates the velocity pixel size for the LSD velocity grid based off the spectral wavelengths.

    Parameters
    ----------
    wavelengths : :py:type:`Array1D`
        Wavelengths for Acid input spectrum (in Angstroms), must be a 1D array of positive values.

    Returns
    -------
    :py:type:`float`
        Velocity pixel size in km/s
    """
    if wavelengths.ndim != 1:
        raise ValueError("Input wavelengths must be a 1D array.")
    if np.any(wavelengths <= 0):
        raise ValueError("Wavelengths must be positive.")
    wavelengths = np.sort(wavelengths)
    return c_kms * np.nanmean(np.diff(np.log(wavelengths)))

@beartype
def guess_SNR(
        frame_wavelengths : Array1D | Array2D,
        frame_flux        : Array1D | Array2D,
        frame_errors      : Array1D | Array2D
        ) -> Array1D | Scalar:
    """Estimates S/N for each frame. Takes the median S/N in the central two-thirds of the
    wavelength range. Fully vectorized so that all the frames can be passed at once.

    Parameters
    ----------
    frame_wavelengths : :py:type:`Array1D | Array2D`
        Array/list of wavelengths for each frame.
    frame_flux : :py:type:`Array1D | Array2D`
        Array/list of flux values for each frame.
    frame_errors : :py:type:`Array1D | Array2D`
        Array/list of error values for each frame.

    Returns
    -------
    :py:type:`Array1D | Scalar`
        Array of estimated signal-to-noise ratios for each frame.
    """
    if np.any(frame_flux) <= 0 or np.any(frame_errors) <= 0 or np.any(frame_wavelengths <= 0):
        raise ValueError("Flux, errors, and wavelengths must all be positive non-zero to estimate S/N.")

    frame_wavelengths = np.atleast_2d(frame_wavelengths)
    frame_flux = np.atleast_2d(frame_flux)
    frame_errors = np.atleast_2d(frame_errors)

    lo, hi = np.nanpercentile(frame_wavelengths, [100/6, 500/6], axis=-1)
    mask = (frame_wavelengths > lo[:, None]) & (frame_wavelengths < hi[:, None])
    frame_flux = np.where(mask, frame_flux, np.nan)
    frame_errors = np.where(mask, frame_errors, np.nan)

    sn_per_pixel = frame_flux / frame_errors
    return np.nanpercentile(sn_per_pixel, 99, axis=-1).squeeze()

@beartype
def guess_errors(
    frame_wavelengths : Array1D | Array2D,
    frame_flux        : Array1D | Array2D,
    frame_sn          : Scalar  | Array1D
    ) -> Array1D | Array2D:
    """
    Estimates errors for each frame based on the flux and S/N.
    Fully vectorized so that all the frames can be passed at once.

    Parameters
    ----------
    frame_wavelengths : :py:type:`Array1D | Array2D`
        Array/list of wavelengths for each frame.
    frame_flux : :py:type:`Array1D | Array2D`
        Array/list of flux values for each frame.
    frame_sn : :py:type:`Scalar | Array1D`
        Estimated signal-to-noise ratio for each frame. Can be a single value or an array of values for each frame.

    Returns
    -------
    :py:type:`Array1D | Array2D`
        Array of estimated error values for each frame. The dimensions will match those of frame_flux, with the same number of frames and pixels.
    """
    if np.any(frame_flux) <= 0 or np.any(frame_wavelengths <= 0) or np.any(frame_sn <= 0):
        raise ValueError("Flux, wavelengths and sn must all be positive non-zero to estimate errors.")

    frame_wavelengths = np.atleast_2d(frame_wavelengths)
    frame_flux = np.atleast_2d(frame_flux)
    frame_sn = np.atleast_1d(frame_sn)
    if frame_sn.ndim > 1 or (frame_sn.ndim == 1 and frame_sn.size != frame_flux.shape[0]):
        raise ValueError("S/N must be a single value or an array of values with the same length as the number of frames in frame_flux.")

    errors = frame_flux / frame_sn[:, None]
    return errors.squeeze()

@beartype
def collapse_SNR(sn:Array1D | Array2D, wavelengths:Array1D | Array2D) -> Array1D|Scalar:
    """
    Collapses the SN of a 1D or 2D wavelength and sn array to the median of the SNs
    on the central 2/3 wavelengths.

    Parameters
    ----------
    sn : :py:type:`Array1D | Array2D`
        Signal-to-noise ratio array.
    wavelengths : :py:type:`Array1D | Array2D`
        Wavelength array.

    Returns
    -------
    :py:type:`Array1D | Scalar`
        Collapsed signal-to-noise ratio.
    """
    sn = np.atleast_2d(sn)
    wavelengths = np.atleast_2d(wavelengths)

    lo, hi = np.nanpercentile(wavelengths, [100/6, 500/6], axis=-1)
    mask = (wavelengths > lo[:, None]) & (wavelengths < hi[:, None])
    return np.nanmedian(np.where(mask, sn, np.nan), axis=-1).squeeze()

@beartype
def get_normalisation_coeffs(wl:Array1D)->tuple[Scalar, Scalar]:
    """Calculates normalization coefficients for wavelength array

    Parameters
    ----------
    wl : array_like
        Wavelength array for which normalization coefficients are calculated.

    Returns
    -------
    tuple
        A tuple containing the normalization coefficients (a, b).
    """
    a = 2 / (np.nanmax(wl)-np.nanmin(wl))
    b = 1 - a * np.nanmax(wl)
    return a, b

def get_available_memory():
    """
    Returns the available memory in bytes.
    Checks if in a SLURM environment and uses its memory allocation if available.

    Returns
    -------
    int
        Available memory in bytes.
    """
    if "SLURM_JOB_ID" in os.environ:
        available_memory = int(os.environ.get('SLURM_MEM_PER_NODE')) # in MB
        available_memory *= 1024**2  # Convert to bytes as in the else statement below
    else:
        available_memory = psutil.virtual_memory().available
    return available_memory

def save_backend_to_hdf5(backend, filename):
    nwalkers, ndim = backend.shape
    niter = backend.iteration

    hdf = emcee.backends.HDFBackend(
        filename,
        dtype=getattr(backend, "dtype", None),
    )

    # Overwrite existing file and reset if already exists
    hdf.reset(nwalkers, ndim)

    blobs = backend.get_blobs()

    # Allocate space in the HDF5 datasets.
    if niter > 0:
        hdf.grow(niter, None if blobs is None else blobs[0])

    with hdf.open("a") as f:
        g = f["mcmc"] # mcmc is always the emcee default

        if niter > 0:
            g["chain"][:] = backend.get_chain()
            g["log_prob"][:] = backend.get_log_prob()
            g["accepted"][:] = backend.accepted

            if blobs is not None:
                g["blobs"][:] = blobs
                g.attrs["has_blobs"] = True

        # Required so HDFBackend.iteration works correctly.
        g.attrs["iteration"] = niter

        # Reload the random state of the walkers if they exist
        random_state = getattr(backend, "random_state", None)
        if random_state is not None:
            for i, v in enumerate(random_state):
                g.attrs[f"random_state_{i}"] = v
    return hdf

def backend_to_sampler(backend, log_prob_fn):
    nwalkers, ndim = backend.shape
    sampler = emcee.EnsembleSampler(nwalkers, ndim, log_prob_fn)
    sampler.backend = backend
    return sampler

def set_dict_defaults(input_dict: dict | None, default_dict: dict) -> dict:
    """Sets default values in a dictionary if they are not already present.

    Parameters
    ----------
    input_dict : dict | None
        The dictionary to set defaults in (or none if not provided).
    default_dict : dict
        The dictionary containing default key-value pairs.
    
    Returns
    -------
    dict
        The updated dictionary with defaults set.
    """
    input_dict = dict(input_dict or {})
    for key, value in default_dict.items():
        input_dict.setdefault(key, value)
    return input_dict

def findfiles(directory:str, file_type:str) -> list[str]:
    """Finds files in a directory matching a specific file type, excluding corrected spectra.

    Parameters
    ----------
    directory : str
        The directory to search for files.
    file_type : str
        The file type to search for.

    Returns
    -------
    list[str]
        A list of files matching the specified file type, excluding corrected spectra.
    """
    filelist_corrected = glob.glob('%s/*/*%s**A_corrected*.fits'%(directory, file_type)) #finding corrected spectra
    filelist = glob.glob('%s/*/*%s**A*.fits'%(directory, file_type)) #finding all A band spectra

    filelist_final = []
    filelist_corrected_set = set(filelist_corrected)

    for file in filelist: #filtering out corrected spectra
        if file not in filelist_corrected_set:
            filelist_final.append(file)

    return filelist_final

def robust_mean(data:np.ndarray, nsig:int|float=3, axis:int=0) -> np.ndarray|float:
    """Calculates the robust mean of the input data by excluding outliers beyond a
    specified number of standard deviations from the median.

    Parameters
    ----------
    data : np.ndarray
        Input data array.
    nsig : int | float, optional
        Number of standard deviations to use for outlier rejection. Defaults to 3.
    axis : int, optional
        Axis along which to compute the robust mean. Defaults to 0.

    Returns
    -------
    float
        Robust mean of the input data.
    """
    median = np.median(data, axis=axis)
    mad = np.median(np.abs(data - median), axis=axis, keepdims=True)
    sigma_nmad = 1.4826 * mad
    mask = np.abs(data - median) < nsig * sigma_nmad
    robust_data = np.where(mask, data, np.nan)
    return np.nanmean(robust_data, axis=axis)

@beartype
def combine_profiles(
    spectra     : Array2D,
    errors      : Array2D | None = None,
    covariances : Array3D | None = None,
    ) -> tuple:
    """
    Combine multiple profiles into one.

    Parameters
    ----------
    spectra : :py:class:`Array2D`
        Input profiles, should have dimensions (n_profiles, n_pix).
    errors : :py:class:`Array2D`, optional
        1-sigma errors for each profile. Used only if covariances is None. Should have dimensions (n_profiles, n_pix).
    covariances : :py:class:`Array3D` or list of :py:class:`Array2D`, optional
        Full covariance matrix for each profile. Should have dimensions (n_profiles, n_pix, n_pix). 
        If provided, this is used instead of errors for the combination, and the combined covariance matrix is also returned.

    Returns
    -------
    combined_profile : :py:class:`Array1D`
    combined_errors : :py:class:`Array1D`
        sqrt(diag(combined_covariance))
    combined_covariance : :py:class:`Array2D`
        Only returned if covariances is provided.
    """
    from scipy.linalg import cho_factor, cho_solve
    spectra = np.asarray(spectra)

    if covariances is not None:
        errors = np.asarray(errors)
        covariances = np.asarray(covariances)

        # We are solving the equations:
        # z_combined ​= (∑​C_i^{−1}*​z_i​) / (∑C_i^−1)
        # And the denominator is the inverse of the combined covariance matrix, so we can get that as well.
        n_profiles, n_pix = spectra.shape
        cho_sum = np.zeros((n_pix, n_pix), dtype=float)
        rhs = np.zeros(n_pix, dtype=float)
        I = np.eye(n_pix)

        for z, C in zip(spectra, covariances):
            cf = cho_factor(C, check_finite=False)
            cho_sum += cho_solve(cf, I, check_finite=False)
            rhs += cho_solve(cf, z, check_finite=False)
        cf_comb = cho_factor(cho_sum, check_finite=False)
        combined_profile = cho_solve(cf_comb, rhs, check_finite=False)
        combined_covariance = cho_solve(cf_comb, I, check_finite=False)
        combined_errors = np.sqrt(np.diag(combined_covariance))
        return combined_profile, combined_errors, combined_covariance

    elif errors is not None:
        errors = np.asarray(errors)

        # Otherwise we do a simple weighted average using the errors, which is equivalent 
        # to the above if the covariance matrices are diagonal.
        weights = 1.0 / errors**2
        combined_profile = np.sum(weights * spectra, axis=0) / np.sum(weights, axis=0)
        combined_errors = np.sqrt(1.0 / np.sum(weights, axis=0))

        return combined_profile, combined_errors

    else:
        combined_profile = np.nanmean(spectra, axis=0)
        return combined_profile

@beartype
def flux_to_od(
    flux       : np.ndarray = None,
    errors     : np.ndarray = None,
    linelist   : np.ndarray = None,
    cov_matrix : np.ndarray = None
    ) -> tuple | np.ndarray:
    """Converts flux, errors, linelist, and covariance matrix to optical depth.
    The formula used for the conversion is:
    - Optical depth (od) is calculated as -log(flux).
    - Errors in optical depth are calculated as errors / flux.
    - Linelist in optical depth is calculated as -log(1 - linelist).
    - Covariance matrix in optical depth is calculated as cov_matrix / (flux[:, np.newaxis] * flux[np.newaxis, :]).

    Parameters
    ----------
    flux : np.ndarray
        Input flux array.
    errors : np.ndarray, optional
        Input errors array. Defaults to None.
    linelist : np.ndarray, optional
        Input linelist array. Defaults to None.
    cov_matrix : np.ndarray, optional
        Input covariance matrix, corresponding to errors. Defaults to None.

    Returns
    -------
    tuple
        A tuple containing any of the following, depending on which inputs were provided:
        - the flux in optical depth
        - errors in optical depth
        - linelist in optical depth
        - covariance matrix in optical depth
        The function appends the above inputs in the respective order to the output tuple if they were input.
    """
    out = []

    if flux is not None:
        flux_od = -np.log(flux)
        out.append(flux_od)
    else:
        flux_od = None

    if errors is not None:
        if flux_od is None:
            raise ValueError("'flux' must be provided if 'errors' is provided.")
        out.append(errors / flux)

    if linelist is not None:
        out.append(-np.log(1 - linelist))

    if cov_matrix is not None:
        if flux_od is None:
            raise ValueError("'flux' must be provided if 'cov_matrix' is provided.")
        out.append(cov_matrix / (flux[:, np.newaxis] * flux[np.newaxis, :]))

    return tuple(out) if len(out) > 1 else out[0]

@beartype
def od_to_flux(
    od         : np.ndarray = None,
    errors     : np.ndarray = None,
    linelist   : np.ndarray = None,
    cov_matrix : np.ndarray = None
    ) -> tuple | np.ndarray:
    """
    Converts optical depth to flux, errors, linelist, and covariance matrix.
    The formula used for the conversion is:
    - Flux is calculated as exp(-od).
    - Errors in flux are calculated as errors * flux.
    - Linelist in flux is calculated as 1 - exp(-linelist).
    - Covariance matrix in flux is calculated as cov_matrix * (flux[:, np.newaxis] * flux[np.newaxis, :]).

    Parameters
    ----------
    od : np.ndarray
        Input optical depth array.
    errors : np.ndarray, optional
        Input errors array. Defaults to None.
    linelist : np.ndarray, optional
        Input linelist array. Defaults to None.
    cov_matrix : np.ndarray, optional
        Input covariance matrix. Defaults to None.

    Returns
    -------
    tuple
        A tuple containing any of the following, depending on which inputs were provided:
        - the flux in flux units
        - errors in flux units
        - linelist in flux units
        - covariance matrix in flux units
        The function appends the above inputs in the respective order to the output tuple if they were input.
    """
    out = []

    if od is not None:
        flux = np.exp(-od)
        out.append(flux)
    else:
        flux = None

    if errors is not None:
        if flux is None:
            raise ValueError("'od' must be provided if 'errors' is provided.")
        out.append(errors * flux)

    if linelist is not None:
        out.append(1-np.exp(-linelist))
    
    if cov_matrix is not None:
        if flux is None:
            raise ValueError("'od' must be provided if 'cov_matrix' is provided.")
        out.append(cov_matrix * (flux[:, np.newaxis] * flux[np.newaxis, :]))

    return tuple(out) if len(out) > 1 else out[0]

def configure_mp_environ(os):
    """
    Configures the multiprocessing environment variables to check for SLURM and set appropriate thread limits.
    See :ref:`multiprocessing` section of the documentation for more details on why this is done.

    Parameters
    ----------
    os : module
        The os module to use for setting environment variables.
    """
    slurm = "SLURM_JOB_ID" in os.environ
    if slurm:
        if os.getenv("OMP_NUM_THREADS") != "1" or os.getenv("MKL_NUM_THREADS") != "1":
            raise ValueError(f"In a SLURM environment, OMP_NUM_THREADS and MKL_NUM_THREADS must be set to 1 before any imports for parallel MCMC. \n" \
            "Please set this in your SLURM job script or at the top of your python script before any other imports.\n" \
            "See https://acid-code.readthedocs.io/en/latest/using_ACID.html#multiprocessing for more information.")
    else:
        # Emcee recommendation, after testing this is absolutely a requirement
        os.environ["OMP_NUM_THREADS"] = "1"
        os.environ["MKL_NUM_THREADS"] = "1"

def next_pow_2(n):
    """
    Calculates the next power of 2 greater than or equal to n.

    Parameters
    ----------
    n : int
        Input number. Must be real and non-negative.

    Returns
    -------
    int
        The next power of 2 greater than or equal to n.
    """
    n = int(n)
    if n < 0:
        raise ValueError("Input must be a non-negative integer.")
    return 1 if n == 0 else 2**(n - 1).bit_length()

def auto_window(taus: np.ndarray, c: float = 5.0):
    """
    Automated windowing procedure following Sokal (1989) in emcee documentation.
    Returns the window index to use in taus[window].
    """
    m = np.arange(len(taus)) < c * taus
    if np.any(m):
        return np.argmin(m)
    return len(taus) - 1

def autocorr_func_1d(x, norm=True):
    """
    Autocorrelation estimate using FFT from the emcee tutorial.
    """
    x = np.atleast_1d(x)
    if len(x.shape) != 1:
        raise ValueError("invalid dimensions for 1D autocorrelation function")
    n = next_pow_2(len(x))

    # Compute the FFT and then (from that) the auto-correlation function
    f = np.fft.fft(x - np.mean(x), n=2 * n)
    acf = np.fft.ifft(f * np.conjugate(f))[: len(x)].real
    acf /= 4 * n

    # Optionally normalize
    if norm:
        acf /= acf[0]

    return acf

def autocorr_gw2010(y, c=5.0):
    """Goodman & Weare (2010) autocorrelation estimate from emcee documentation"""
    f = autocorr_func_1d(np.mean(y, axis=0))
    taus = 2.0 * np.cumsum(f) - 1.0
    window = auto_window(taus, c)
    return taus[window]

def autocorr_new(y, c=5.0):
    """New integrated autocorrelation time estimate from emcee documentation"""
    
    # Average ACF across walkers
    assert y.ndim == 2, "Expects y with shape (nwalkers, nsteps)"
    nwalkers, nsteps = y.shape

    f = np.zeros(nsteps)
    for walker in range(nwalkers):
        f += autocorr_func_1d(y[walker], norm=True)
    f /= nwalkers

    # Apply sokal windowing on cumulative sum
    taus = 2.0 * np.cumsum(f) - 1.0
    window = auto_window(taus, c)
    return float(taus[window])

def sampler_nbytes(sampler) -> IntLike:
    if hasattr(sampler, "nbytes"):
        return sampler.nbytes
    elif hasattr(sampler.backend, "nbytes"):
        return sampler.backend.nbytes
    else:
        nbytes = 0
        for name in ("chain", "log_prob", "accepted", "blobs"):
            arr = getattr(sampler.backend, name, None)
            if arr is not None and hasattr(arr, "nbytes"):
                nbytes += arr.nbytes
        return nbytes
